
/* mpeg2dec.c, main(), initialization, option processing                    */

/* Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The MPEG Software Simulation Group does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 *
 * Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
 * are subject to royalty fees to patent holders.  Many of these patents are
 * general enough such that they are unavoidable regardless of implementation
 * design.
 *
 * MPEG2AVI REVISION HISTORY
 * -------------------------
 *   0.16B35 increased memory padding on calloc() calls
 *      allow IVTC to be initialized to ODD polarity.
 *      -s0 through -s4 are even polarity, -s5 through -s9 are odd
 *      the ivtc routine converts s5-s9 back to s0-s4
 *   0.16B34 declared unsigned char g_scan[2][64], which the MPEG-decoder uses
 *      to scan macroblock coefficiens.  g_scan[] is initialized during
 *      run-time, so the decoder can run with both transposing and 
 *      non-transposing iDCT functions efficiently.
 *      fixed bug with default iDCT mode : now it really is _IDCT_MMX32 (doh!)
 *      converted malloc() calls to calloc() calls
 *      removed int32_idct entirely
 *   0.16B33 more mmx optimization to downsizer, restructuring of vresizer.c
 *      all motino-compensation PEL cases use aligned reads (recon.c)
 *      modified MMX AAN IDCT - transpose is now integrated
 *      Chen MMX IDCT hand-scheduled for better performance on Pentium/MMX
 *      replaced INT32-iDCT ('-r1') with MXM32-IDCT (IEEE-1180 compliant)
 *       (note, due to my error, 0.16B33 defaults to INT32-iDCT!)
 *      removed the demux_queue from MPEG2AVI
 *      changed INT32 iDCT to ('-r4'), still available to user
 *   0.16B32 mmx optimizations to the resizers CHROMA420 Y-downsizers.  
 *      Scaling in the Y-direction should not be somewhat faster.  The resizer
 *      now imposes a minimum source/destination pixel-size of 32, to make
 *      sure the internal scaling coefficients fit inside a 15-bit magnitude.
 *   0.16B31 resizer functions shrink() [vresizer.c] minor optimization
 *      force user to specify '-f' option, to fix AVI-encoding problems with
 *      coded interlaced sequences (no repeat-first field),
 *      some further optimizations to recon.c (read alignment for reads)
 *  10/15/99
 *   0.16B3  changed default for UserOptVOBStartLBA to 0, to always engage
 *      seek mode
 *      made the printed-information from '-*' more accurate, hopefully
 *       frame-accurate
 *      replaced getchar(), _getch() calls with scanf().
 *      encapsulated the streamlist feature into a C++ object cStreamlist
 *  10/10/99
 *   0.16B24 added my_open(), my_read(), and my_close() to mpv_read.cpp
 *      all low-level IO accesses are routed through the my_xxx() functions
 *      this allows continuous-stream reading
 *
 *  10/08/99
 *   0.16B23 more MMX-optimizations (DWORD aligned -> Add_Block() getpic.c)
 *          MMX color-space conversion (very slightly less accurate than int)
 *          moved yuv444torgb24() to new file, yuv2rgb.c
 *          moved yuv(planar) -> yuv(packed) to yuv2yuv.c
 *           (these functions were integrated into a switch() in store.c )
 *          store.c - prevent user from writing-AVI with _IDCT_MMX_AAN 

 *  10/6/99
 *   0.16B22 MMX optimizations, added test_cpu() function to test for MMX.
 *          recon.c is fully MMX-optimized (all motion-compensation cases
 *          run MMX asm code.  Not necessarily the most efficient code...)
 *          getpic.c 's Add_Block() and Clear_Block() are MMX-optimized
 *          most frequently used fast-filters (downsample_half(), yuv2yuv)
 *           are MMX-optimized
 *          changed default filter-mode to '-q1' (fast-filter), since most
 *           fast-filters are MMX-optimized
 *          Initialize all IDCT algorithms
 *          changed meaning of '-r', now user chooses IDCT-function with
 *           '-r<n>', where <n> is chosen from a table (in global.h)
 *           If the table is modified, store.c and mpeg2dec.c must be changed

 *
 *   0.16B2 added UserOptPrintFrameCode, prints <MPEG frame#> in upper-left 
 *          corner of the output frame.  option '-*' for video watermarking
 *          '-@ <lba>' VOB LBA seek, target must be SYSTEM HEADER packet!
 *          enhanced input option '-b' with '-b filename [stream_id]',
 *           [stream_id] = optional StreamID (hex) parameter
 *          example '-b input.vob 0xE1'
 *          BBDMUX has been upgraded to v1.2 and converted to C++ (cDemuxer)
 *  09/20/99
 *   0.16B1 moved the AVI-backend to C++,
 *          added '-$ <MB>' UserOptAVISplitLength to support auto AVI-splitting
 *          added '-! <n>' UserOptAVISkipEvery to drop 1 MPEG-frame per <n> 
 *            MPEG-frames, the frame-drop is done at the last possible second
 *            (to not disturb IVTC)
 *      
 *  09/14/99
 *   0.15B51(finally) fixed AVI-YUV modes and video-codec setup, see
 *          storeavi.c for details
 *          also removed "2X pixel-granularity" restrictions on shrink420()
 *          and shrink422() functions.  The restriction was totally 
 *          unnecessary, due to a mathematical misperception on my part.
 *          Added -# option for frame-start/end points,
 *          g_cmd_buttonStop, g_cbMPEGPicture (gethdr.c) vars added
 *           -# is implemented in two places, gethdr.c and store.c
 *           the FRangeStart is implemented in gethdr.c

 *  09/11/99
 *   0.155 restructured the AVI/GDI output dataflow.  Now, the output images
 *          are resized, then (optionally) halved, and finally converted to
 *          the final image format.
 *          | MPEG decoder | -> | resizer | -> | half-res | -> | Crop & CSC |
 *               store.c     ->  shrink_frame()-> yuv2yuv() ->  | yuv2rgb24()
 *          added YUY2 output mode
 *          
 *  09/04/99  
 *   0.154 corrected bug with fast-filter (-q1) when decoding MPEG-1 video
 *          reduced #references to  global variables in vidfilt.c, vresizer.c
 *          +converted g_AVIcrop -> UserOptCrop
 *
 *  08/15/99  v0.153
 *   0.153 further performance tweaks to resizer.c, vidfilt.c
 *          +auto-disable IVTC feature (gethdr.c, telecine.c)
 *          +converted g_ivtc_start_state -> struct UserOptIVTC
 *   0.152 moved resizer functions shrink888() to vresize.c, added
 *          shrink888xy() for faster downsampling when both X/Y are adjusted.
 *
 *   0.15 added fast/Hi-Q filter control '-q<n>', fast-upscaling for YUV420->YUV444
 *        deafult is Hi-Q filter ( -q0 )
 *        changed all references from open()/close()/write() to 
 *        _open(), _write(), _close()
 *        (low-pass) downsampling filter for half-resolution mode
 *        new file : vidfilt.c (upsample/downsample filters moved here)
 *        created new type ( struct tUserOption ) to manage/transport user options
 *		  between the main-parser (mpeg2dec.c) and decoder/output routines.
 *        (see useropt.h)
 *
 *  08/11/99  v0.14
 *   0.14 added 1/2-resolution support and cropping to output modes -o6, -o7
 *        added necessary variables g_AVIcropX, g_AVIcropY, g_vHalfResFlag
 *              Resized_Picture_Width, Resized_Picture_Height
 *        removed automatic "-r" (FP-IDCT) with "-o7" usage
 *        tweaked ivtc algorithms, fixed "color bleed" problem in full
 *        resolution mode.
 *
 *  08/09/99  v0.13
 *   0.13 preliminary support for VOB files using BBDMUX,
 *        abort key 'Q' to terminate AVI encoding (while preserving AVI file)
 *        BBDMUX is the work of Brent Beyeler (http://members.home.net/beyeler)
 *
 *  08/06/99  v0.12
 *   0.12 user-selected frame-rate code
 *   0.11 corrected AVI [uncompressed] write mode, incorrect header size 
 *   0.10 first public release
 *  modifications by Royce Shih-Wea Liao (liaor@iname.com) 
 *
 ***************************************************************
 *  mpeg2dec.c (modified)
 *		added option to display images (as 24bpp RGB) to Windows GDI
 *		added option to perform inverse-telecine (inverse 3:2 pulldown)
 *  
 *  aviwrite.c (C++ object)
 *      starting with v0.16B1, the AVI-writer is now encapsulated as a C++ 
 *      object.  There is an additional file "aviinit.c" that acts like
 *      "glue logic", to stitch store_avi_gdi() [store.c] to the C++
 *      object.  Not very elegant, but simplifies future updates to
 *      the AVIwriter.
 *
 *  windisp.c (new file)
 *		draw_dib() routine
 *      implements -o6
 *
 *  telecine.c (new file)
 *		inverse_telecine() routine
 *      ivtc_full_8bpp()
 *      ivtc_half_8bpp()
 *      get_ivtc_cmd()
 *
 *  store.c (modified)
 *      rstrcmpi()
 *      store_avi_gdi() -  handles -o6 and -o7 functions
 *      supports inverse_telecine() for -o6 and -o7 only!
 *
 *  storeavi.c (new file)
 *      implements -o7
 *
 *  useropt.h (new file)
 *		implements <struct tUserOption>
 *      this structure encapsulates "UserOptions", i.e. command-line parameters
 *      It's gradual phase-in will hopefully limit the # "free floating" global 
 *      variables
 *		CLEAROPT( tUserOption );  // quick-macro : clears tUserOption
 *		if ( CHECKOPT( tUserOption ) ) // quick-macro 
 *
 *
 *  vidfilt.c (new file)
 *      all u420to422, u422to444, u444to422, u422to420, 
 *      The command-var (g_cmd_filter)
 *      is processed in here.  Color-space convertors are stored here 
 *      as well.
 *
 *  vresizer.c (new file )   aspect-ratio (downsizing) functions 
 *      shrink888x(), shrink888y(), shrink888xy() 
 *      downsample_half(), half420to420()
 *
 *  vtitler.c (new file)  print text onto the video (very primitive, #s only)
 *
 *  yuv2yuv.c (new file) convert MPEG-format planar pixel format (420/422/444)
 *      to a different format (420/422/444.)  Any input-type can be convertrd
 *      to any output-type.  Additionaly, "half-resolution" option auto
 *      downsamples the input in an optimal manner (both performance-wise and
 *      quality-wise.)
 *
 *  demuxbuf.cpp - a C++ queue that sits between the cDemuxer [bbdmux12.h]
 *      and s_read.c [ fread() emulator for MPEG2Decoder ]
 *  bbdmux12.cpp, bbdbits.cpp (.h), demuxbuf.cpp (.h), s_read.cpp
 *      together, these files implement stream-demuxing through BBDMUX
 *      MPEG-2 video (stream 0xE0 only) can be extracted from VOB files
 *      s_read.cpp is a "wrapper" for getbits.c's ( _read() )
 *      demuxbuf.cpp is a C++ class-definition which implements a queue
 *      to connect BBDMUX's mpeg2_demux() and MPEG2V12's _read() 
 *  idct_aan.c, idctchen.c
 *      16-bit MMX implementations of the IDCT operation.  Neither MMX IDCT
 *      is IEEE-1180 compliant.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <io.h>
#include <conio.h>

#define GLOBAL
#include "config.h"
#include "global.h"
#include "asm.h"
#include <math.h>

#include <windows.h>

#define EXTRA_MEM_PAD 16	// extra bytes to pad each framebuffer
	// 0.16B22  EXTRA_MEM_PAD allows MMX-operations to read beyond the
	// logical end of a video-frame (for example, an MMX-loop might
	// right up to 7 bytes beyond the end of the frame, for DWORD/QWORD
	// alignment purposes.)


// v0.16B24, these functions are located in mpv_read.cpp
// they are manually declared in this file, as extern, because the
// functions are located in a C++ file.  I cannot include the
// function definition in the global.h file because the global.h
// is used by both the C++ files and C files.  The linker becomes
// confused and thinks mpv_open(), mpv_close() are C++ functions 
extern int mpv_open( const char *filename, int oflag, const char* outDir);
extern int mpv_close( int handle );
//extern long mpv_lseek( int handle, long offset, int origin );


/* private function prototypes, and private variables */
static int  video_sequence _ANSI_ARGS_((int *framenum));
static int Decode_Bitstream _ANSI_ARGS_((void));
static int  Headers _ANSI_ARGS_((void));
static void Initialize_Sequence _ANSI_ARGS_((void));
static void Initialize_Decoder _ANSI_ARGS_((void));
static void Deinitialize_Sequence _ANSI_ARGS_((void));
//PX3 static void Process_Options _ANSI_ARGS_((int argc, char *argv[]));
	// test CPU for MMX-capability
static void test_cpu( BOOL *p_fMMX, BOOL *p_f3DNOW );	
	// v0.16b34 initialize scan order matrix g_scan[2][64]
static void init_scan_matrix( int type );

// for init_scan_matrix()
#define SCAN_ORDER_TRANSPOSE 1
#define SCAN_ORDER_NORMAL    0

#if OLD
static int  Get_Val _ANSI_ARGS_((char *argv[]));
#endif

/* #define DEBUG */

static void Clear_Options();
#ifdef DEBUG
static void Print_Options();
#endif 



BOOL m_fMMX , m_f3DNOW; /* flags for CPU capabilities */

void *fTempArray, *p_fTempArray;

void InitEngine( void )
{
	DWORD dwrc = 0; /* dword return code */

	m_fMMX = FALSE;
	m_f3DNOW = FALSE; /* flags for CPU capabilities */
	
	Clear_Options();

	/* 0.16B22 check for MMX capability */
	test_cpu( &m_fMMX, &m_f3DNOW );

	if ( !m_fMMX )
	{
		fprintf( stderr, "\nmain() : Your CPU does *NOT* support MMX instructions!");
		fprintf( stderr, "\n\tThis version of <%s> requires MMX support!", Version );
		exit( ERROR );
	}
}


/* Main entry point for encoding via lib */
int EncodeMpeg2Avi( void )
{

	int ret, code;


	/* v0.16B31 enforce '-f' */
	if ( !Frame_Store_Flag )
	{
		fprintf( stderr, "\nFrame_Store_Flag is not specified." );
		fprintf( stderr, "\nYou must include '-f' in your command line!" );
		exit( ERROR );
	}
	
	// the input file is a streamlist, automatically engage BBDMUX
	
	if ( rstrcmpi( Main_Bitstream_Filename, ".vob" ) == 0 ||
		rstrcmpi( Main_Bitstream_Filename, ".lst" ) == 0  )
		g_vobflag = 1; /* DVD vob file, use BBDMUX */
	else
		g_vobflag = 0; /* M2V file, don't use BBDMUX */
	
	
	ld = &base; /* select base layer context */
	
	/* open MPEG base layer bitstream file(s) */
	/* NOTE: this is either a base layer stream or a spatial enhancement stream */
	
	/* if g_vobflag is set, then the routines in s_read.cpp will automatically
	"connect" to BBDMUX, and BBDMUX will perform an fopen() on the filename */
	
	//  if ( !g_vobflag )
	//  {  
    if ((base.bs_Infile=mpv_open(Main_Bitstream_Filename,O_RDONLY|O_BINARY, Output_Directory))<0)
	{
		fprintf(stderr,"\nBase layer input file '%s' not found", Main_Bitstream_Filename);
		exit(1);
	}
	
	//  if ( g_vobflag )
	if ( !g_vobflag )
	{
	/*	  printf("\nUsing BBDMUX to decode bitstream 0xE0 from %s", 
		  Main_Bitstream_Filename );
		  
			if ( Frame_Store_Flag == 0 )
			{
			fprintf(stderr,"Warning : -f option was not specified!");
			fprintf(stderr,"\n        (Post-processing functions may fail!)");
			}
	}*/
		
		
		if(base.bs_Infile != 0)
		{
			Initialize_Buffer(); 
			
			if(Show_Bits(8)==0x47)
			{
				sprintf(Error_Text,"Decoder currently does not parse transport streams\n");
				Error(Error_Text);
			}
			
			next_start_code();
			code = Show_Bits(32);
			
			switch(code)
			{
			case SEQUENCE_HEADER_CODE:
				break;
			case PACK_START_CODE:
				System_Stream_Flag = 1;
			case VIDEO_ELEMENTARY_STREAM:
				System_Stream_Flag = 1;
				break;
			default:
				sprintf(Error_Text,"Unable to recognize stream type\n");
				Error(Error_Text);
				break;
			}
			
			// v0.16B35 instead of using mpv_lseek(), close and re-open the file
			mpv_close(base.bs_Infile);
			
			if ((base.bs_Infile=mpv_open(Main_Bitstream_Filename,O_RDONLY|O_BINARY, Output_Directory))<0)
			{
				fprintf(stderr,"\nBase layer input file '%s' could not be re-opened", Main_Bitstream_Filename);
				exit(1);
			}
			
			//  mpv_lseek(base.bs_Infile, 0l, SEEK_SET); 
		} /* endif ( base.bs_Infile != 0) */
	} /* endif -- if ( !g_vobflag ) */
	
	Initialize_Buffer(); 
	
	/* disable two_streams support
	if(Two_Streams)
	{
    ld = &enhan;  select enhancement layer context 
	
	  if ((enhan.Infile = _open(Enhancement_Layer_Bitstream_Filename,O_RDONLY|O_BINARY))<0)
	  {
      sprintf(Error_Text,"enhancment layer bitstream file %s not found\n",
	  Enhancement_Layer_Bitstream_Filename);
	  
		Error(Error_Text);
		}
		
		  Initialize_Buffer();
		  ld = &base;
		  }
	*/
	
	Initialize_Decoder();
	if ( Output_Type == T_AVI || Output_Type == T_AVI_YV12 ||
		Output_Type == T_AVI_YUY2 )
	{
		;
	} // endif
	
	ret = Decode_Bitstream();
	mpv_close(base.bs_Infile);
	
	
	/*if (Two_Streams)
    _close(enhan.Infile);*/
	
	/* rsl */
	if ( Output_Type == T_AVI || Output_Type == T_AVI_YV12 
		|| Output_Type == T_AVI_YUY2 )
	{
		if ( Output_Picture_Filename != NULL )
		{
			printf( "\nClosing output (AVI) file %s", Output_Picture_Filename );
			avi_fileclose();
		}
	}
	
	if ( Output_Type == T_GDI )
	{
		printf("\r<Done.>");
		_getch();
	}


	return 0;
}



/* IMPLEMENTAION specific rouintes */
static void Initialize_Decoder()
{
  int i;

  /* Clip table */
  //if (!(Clip=(unsigned char *)malloc(1024)))
  if (!(Clip=(unsigned char *)calloc(32, 32)))
    Error("Clip[] malloc failed\n");

  Clip += 384;

  for (i=-384; i<640; i++)
    Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  /* IDCT */
  /* v0.16B22 initialize all IDCTs */
//  if (Reference_IDCT_Flag)
    init_idct_ref64();
//  else
//  init_idct_int32(); 

	// v0.16B22 set (*p_idctFunction)() to an IDCT-function.
	// The switch statement parses the user's IDCT-selection
	//
	// The idct-function is called with this statement :
	//  short *block; // IDCT block data (8x8 elements)
	//
	//   (*p_idctFunction)( block );

	switch ( GETOPT( UserOptIDCT ) )
	{
		case _IDCT_FPU64 :
			p_idctFunction = idct_ref64;
			init_scan_matrix( SCAN_ORDER_NORMAL );
			break;		
		case _IDCT_MMX32 :
			p_idctFunction = idct_mmx32; 
			init_scan_matrix( SCAN_ORDER_NORMAL );//SCAN_ORDER_TRANSPOSE ); 
			break;
		case _IDCT_MMX_CHEN1 :
			p_idctFunction = idct_mmx_chen1;
			init_scan_matrix( SCAN_ORDER_NORMAL );
			break;
		case _IDCT_MMX_AAN :
			p_idctFunction = idct_mmx_aan;
			init_scan_matrix( SCAN_ORDER_TRANSPOSE );
			break;
		case _IDCT_AMD_922_AMD:
			p_idctFunction = AMD_IDCT; 
			p_fTempArray = (void *)malloc(sizeof(float)*128 + 64);
		    fTempArray = (void *)((long)p_fTempArray + 64 - (long)p_fTempArray%64);
			init_scan_matrix( SCAN_ORDER_NORMAL );
			break;
		case _IDCT_AMD_922_SSE:
			p_idctFunction = SSE_IDCT; 
			p_fTempArray = (void *)malloc(sizeof(float)*128 + 64);
		    fTempArray = (void *)((long)p_fTempArray + 64 - (long)p_fTempArray%64);
			init_scan_matrix( SCAN_ORDER_NORMAL );
			break;
		case _IDCT_FAST_MIHA: 
			p_idctFunction = Fast_IDCT;
			Initialize_Fast_IDCT();
			init_scan_matrix( SCAN_ORDER_NORMAL );
			break;
		case _IDCT_REF_MIHA: 
			p_idctFunction = REF_IDCT;
			Initialize_REF_IDCT();
			init_scan_matrix( SCAN_ORDER_NORMAL );
			break;
		default :
			p_idctFunction = idct_mmx_chen1;
			init_scan_matrix( SCAN_ORDER_NORMAL );
			break;
	}
}

/* mostly IMPLEMENTAION specific rouintes */
static void Initialize_Sequence()
{
  int cc, size;
  static int Table_6_20[3] = {6,8,12};

  /* check scalability mode of enhancement layer */
  if (Two_Streams && (enhan.scalable_mode!=SC_SNR) && (base.scalable_mode!=SC_DP))
    Error("unsupported scalability mode\n");

  /* force MPEG-1 parameters for proper decoder behavior */
  /* see ISO/IEC 13818-2 section D.9.14 */
  if (!base.MPEG2_Flag)
  {
    progressive_sequence = 1;
    progressive_frame = 1;
    picture_structure = FRAME_PICTURE;
    frame_pred_frame_dct = 1;
    chroma_format = CHROMA420;
    matrix_coefficients = 5;
  }

  /* round to nearest multiple of coded macroblocks */
  /* ISO/IEC 13818-2 section 6.3.3 sequence_header() */
  mb_width = (horizontal_size+15)/16;
  mb_height = (base.MPEG2_Flag && !progressive_sequence) ? 2*((vertical_size+31)/32)
                                        : (vertical_size+15)/16;

  Coded_Picture_Width = 16*mb_width;
  Coded_Picture_Height = 16*mb_height;

  /* ISO/IEC 13818-2 sections 6.1.1.8, 6.1.1.9, and 6.1.1.10 */
  Chroma_Width = (chroma_format==CHROMA444) ? Coded_Picture_Width
                                           : Coded_Picture_Width>>1;
  Chroma_Height = (chroma_format!=CHROMA420) ? Coded_Picture_Height
                                            : Coded_Picture_Height>>1;
  
  /* derived based on Table 6-20 in ISO/IEC 13818-2 section 6.3.17 */
  block_count = Table_6_20[chroma_format-1];

  // v0.16B23, to support aligned qword reading in Add_Block() [getpic.c],
  // the framebuffer sizes must be extended an extra few bytes (EXTRA_MEM_PAD.)
  // This allows the MMX-code to read past the end of the frame, without 
  // causing a protection-fault

  for (cc=0; cc<3; cc++)
  {
    if (cc==0)
      size = (Coded_Picture_Width*Coded_Picture_Height) + EXTRA_MEM_PAD;
    else
      size = Chroma_Width*Chroma_Height + EXTRA_MEM_PAD;

	if (!(backward_reference_frame[cc] = (unsigned char *)calloc( (size>>5)+2,32)))
      Error("backward_reference_frame[] malloc failed\n");

    if (!(forward_reference_frame[cc] = (unsigned char *)calloc( (size>>5)+2,32)))
      Error("forward_reference_frame[] malloc failed\n");

    if (!(auxframe[cc] = (unsigned char *)calloc( (size>>5)+2,32)))
      Error("auxframe[] malloc failed\n");

    if(Ersatz_Flag)
      if (!(substitute_frame[cc] = (unsigned char *)calloc( (size>>5)+2,32)))
        Error("substitute_frame[] malloc failed\n");


//snr    if (base.scalable_mode==SC_SPAT)
//snr    {
//snr      /* this assumes lower layer is 4:2:0 */
//snr      if (!(llframe0[cc] = (unsigned char *)malloc((lower_layer_prediction_horizontal_size*lower_layer_prediction_vertical_size)/(cc?4:1))))
//snr        Error("llframe0 malloc failed\n");
//snr      if (!(llframe1[cc] = (unsigned char *)malloc((lower_layer_prediction_horizontal_size*lower_layer_prediction_vertical_size)/(cc?4:1))))
//snr        Error("llframe1 malloc failed\n");
//snr    }
  }

  /* SCALABILITY: Spatial */
//snr  if (base.scalable_mode==SC_SPAT)
//snr  {
//snr    if (!(lltmp = (short *)malloc(lower_layer_prediction_horizontal_size*((lower_layer_prediction_vertical_size*vertical_subsampling_factor_n)/vertical_subsampling_factor_m)*sizeof(short))))
//snr      Error("lltmp malloc failed\n");
//snr  }

#ifdef DISPLAY
  if (Output_Type==T_X11)
  {
    Initialize_Display_Process("");
    Initialize_Dither_Matrix();
  }
#endif /* DISPLAY */

}

void Error(char* text)
{
  fprintf(stderr,text);
  exit(1);
}

/* Trace_Flag output */
void Print_Bits(int code, int bits, int len)
{
  int i;
  for (i=0; i<len; i++)
    printf("%d",(code>>(bits-1-i))&1);
}







static int Headers()
{
  int ret;

  ld = &base;
  

  /* return when end of sequence (0) or picture
     header has been parsed (1) */

  ret = Get_Hdr();


  if (Two_Streams)
  {
    ld = &enhan;
    if (Get_Hdr()!=ret && !Quiet_Flag)
      fprintf(stderr,"streams out of sync\n");
    ld = &base;
  }

  return ret;
}



static int Decode_Bitstream()
{
  int ret;
  int Bitstream_Framenum;



  Bitstream_Framenum = 0;

  for(;;)
  {

#ifdef VERIFY
    Clear_Verify_Headers();
#endif /* VERIFY */

    ret = Headers();
    
    if(ret==1)
    {
      ret = video_sequence(&Bitstream_Framenum);
    }
    else
      return(ret);

  }

}


static void Deinitialize_Sequence()
{
  int i;

  /* clear flags */
  base.MPEG2_Flag=0;

  for(i=0;i<3;i++)
  {
    free(backward_reference_frame[i]);
    free(forward_reference_frame[i]);
    free(auxframe[i]);

    if (base.scalable_mode==SC_SPAT)
    {
     free(llframe0[i]);
     free(llframe1[i]);
    }
  }

  if (base.scalable_mode==SC_SPAT)
    free(lltmp);

#ifdef DISPLAY
  if (Output_Type==T_X11) 
    Terminate_Display_Process();
#endif
}



static void TimeCalc()
{
	static LDiff = 0;
	static time_t CurrentTime; 
	static long   DiffTime; 
	static long R;
	static int  hours, minutes, seconds;
	static double a,b;

	if ( st_RunState.filePos == 0 )
		return;

	// secure emms PX3 ver 0.1.0
	__asm emms;

	time(&CurrentTime); 

	DiffTime = (long)difftime(CurrentTime,st_RunState.startTime);

	if ( DiffTime == LDiff )
		return;

	st_RunState.FPS = ( (double)st_RunState.encFrames / (double)DiffTime );
				   
	LDiff = DiffTime;


	a = ((st_RunState.fileSize)-(st_RunState.filePos))/10000;
	b = (st_RunState.filePos)/10000;
	R = (long)((double)(a * DiffTime) / b);

	seconds  = (long)R;
	seconds  = seconds%60;
	minutes  = (int)floor((double)R/60);
	minutes  = minutes%60;
	hours    = (int)floor((double)R/3600);

	sprintf( st_RunState.RemainTime, "%02d:%02d:%02d",
		     hours,minutes,seconds);

	seconds  = (long)LDiff;
	seconds  = seconds%60;
	minutes  = (int)floor((double)LDiff/60);
	minutes  = minutes%60;
	hours    = (int)floor((double)LDiff/3600);

	sprintf( st_RunState.ElapsedTime, "%02d:%02d:%02d",
		     hours,minutes,seconds);
}


static int video_sequence(int *Bitstream_Framenumber)
{
  int Bitstream_Framenum;
  int Sequence_Framenum;
  int Return_Value;
  short dispout;
  static int first = 1;

  Bitstream_Framenum = *Bitstream_Framenumber;
  Sequence_Framenum=0;

  Initialize_Sequence();

  /* decode picture whose header has already been parsed in 
     Decode_Bitstream() */


  Decode_Picture(Bitstream_Framenum, Sequence_Framenum);

  /* update picture numbers */
  if (!Second_Field)
  {
    Bitstream_Framenum++;
    Sequence_Framenum++;
  }

  dispout = 0;

  /* loop through the rest of the pictures in the sequence */
  while ((Return_Value=Headers()))
  {

    Decode_Picture(Bitstream_Framenum, Sequence_Framenum);

    if (!Second_Field)
    {
      Bitstream_Framenum++;
      Sequence_Framenum++;
    }


	if ( Output_Type != 6 )
	{
		TimeCalc();
		printf("\rFrame#%lu : AVI (%.0f KB) #KF=%lu #FPS=%.2f #ET=%s #RT=%s ",
			st_RunState.encFrames,
			st_RunState.kbWritten,
			st_RunState.keyFrames,
			st_RunState.FPS,
			st_RunState.ElapsedTime,
			st_RunState.RemainTime );
	} 
	else
	{
		printf("\rFrame#%lu", st_RunState.encFrames );
	}

  }

  /* put last frame */
  if (Sequence_Framenum!=0)
  {
    Output_Last_Frame_of_Sequence(Bitstream_Framenum);
  }

  Deinitialize_Sequence();

#ifdef VERIFY
    Clear_Verify_Headers();
#endif /* VERIFY */

  *Bitstream_Framenumber = Bitstream_Framenum;
  return(Return_Value);
}



static void Clear_Options()
{
  Verbose_Flag = 0;
  Output_Type = -1;
  Output_Picture_Filename[0] = 0;
  Output_Directory[0] = 0;
  hiQdither  = 0;
  Output_Type = 0;
  Frame_Store_Flag = 0;
  Spatial_Flag = 0;
  Lower_Layer_Picture_Filename = " ";
  Reference_IDCT_Flag = 0;
  Trace_Flag = 0;
  Quiet_Flag = 0;
  Ersatz_Flag = 0;
  Substitute_Picture_Filename  = " ";
  Two_Streams = 0;
  Enhancement_Layer_Bitstream_Filename = " ";
  Big_Picture_Flag = 0;
  Main_Bitstream_Flag = 0;
  Main_Bitstream_Filename = " ";
  Verify_Flag = 0; 
  Stats_Flag  = 0;
  User_Data_Flag = 0; 
  g_vHalfResFlag = 0; /* output pictures are downsampled to 1/2 resolution */
  /*g_AVIcropX= _DISABLE_CROP;  for AVI/GDI output */
  /*g_AVIcropY= _DISABLE_CROP;  size of the encoded AVI frames */
  g_vobflag = 0; /* don't use BBDMUX, read-input bitstream directly  */
  Resized_Picture_Width = 0;
  Resized_Picture_Height = 0;
	/* will be set later by store.c, if -o6 or -o7 is active */
  g_cmd_ivtc = 0; /* command-variable for IVTC, see telecine.c */
  g_cmd_filter = 0; /* command-variable for up/down filtering, see store.c */
  g_cmd_buttonStop = 0; /* command-variable for stopping the decoder */

  g_fastfilter = 1; /* 0.16B22, now fast-filter is ON by default */
  g_cbMPEGPicture = 0; /* reset MPEG-picture counter to 0 */
  g_dwSystemPackLBA = -1; /* reset MPEG-System Header LBA position */

  CLEAROPT( UserOptIVTC );    /* clear user-option fields */
  CLEAROPT( UserOptResizeX );
  CLEAROPT( UserOptResizeY );
  CLEAROPT( UserOptCropX );
  CLEAROPT( UserOptCropY );
  CLEAROPT( UserOptFRangeStart );
  CLEAROPT( UserOptFRangeEnd );
  CLEAROPT( UserOptAVIrate );
  
  /* 0.16B1 AVI file splitting, AVIwriter automatically closes and opens new 
	file after #bytes specified by this option */
  CLEAROPT( UserOptAVISplitLength );
  /* if user does not specify, auto-splitting after ~2000MB */
  SETOPT( UserOptAVISplitLength, 1024*1024*2000 );

  CLEAROPT( UserOptAVISkipEvery ); /* 0.16B1 */
  CLEAROPT( UserOptPrintFrameCode ); /* 0.16B2 */
  CLEAROPT( UserOptVOBStartLBA ); /* 0.16B2 */
  SETOPT( UserOptVOBStartLBA, 0 ); /* 0.16B3, always scan VOB for sysheader */
  CLEAROPT( UserOptVOBStreamID ); /* 0.16B2 */

  CLEAROPT( UserOptIDCT ); /* 0.16B22 */
  SETOPT( UserOptIDCT, _IDCT_MMX32 ); /* v0.16B34 default IDCT = 32-bit MMX */

  CLEAROPT( UserOptNoAudioDecoding );
  SETOPT( UserOptNoAudioDecoding, 2 ); /* Default decode all audio */

  CLEAROPT( UserOptAutoResize ); /* PX3 Ver. 0.1.0 */
  SETOPT( UserOptAutoResize, 0 ); /* Autoresizing in preview mode */

  p_idctFunction = NULL; // reset IDCT function pointer to NULL

  st_RunState.fileCount = 0;
  st_RunState.curFile[0] = '\0';
  st_RunState.fileSize = 0;
  st_RunState.filePos = 0;
  st_RunState.startTime = 0;
  st_RunState.encFrames = 0;
  st_RunState.kbWritten = 0;
  st_RunState.keyFrames = 0;
  st_RunState.FPS = 0;
  st_RunState.lowestFPS = 0;
  st_RunState.highestFPS = 0;

  st_RunState.RemainTime[0] = '\0';
  strcpy( st_RunState.RemainTime, "00:00:00" );

  st_RunState.ElapsedTime[0] = '\0';
  strcpy( st_RunState.ElapsedTime, "00:00:00" );

}



#ifdef DEBUG
static void Print_Options()
{
  
  printf("Verbose_Flag                         = %d\n", Verbose_Flag);
  printf("Output_Type                          = %d\n", Output_Type);
  printf("Output_Picture_Filename              = %s\n", Output_Picture_Filename);
  printf("hiQdither                            = %d\n", hiQdither);
  printf("Output_Type                          = %d\n", Output_Type);
  printf("Frame_Store_Flag                     = %d\n", Frame_Store_Flag);
  printf("Spatial_Flag                         = %d\n", Spatial_Flag);
  printf("Lower_Layer_Picture_Filename         = %s\n", Lower_Layer_Picture_Filename);
  printf("Reference_IDCT_Flag                  = %d\n", Reference_IDCT_Flag);
  printf("Trace_Flag                           = %d\n", Trace_Flag);
  printf("Quiet_Flag                           = %d\n", Quiet_Flag);
  printf("Ersatz_Flag                          = %d\n", Ersatz_Flag);
  printf("Substitute_Picture_Filename          = %s\n", Substitute_Picture_Filename);
  printf("Two_Streams                          = %d\n", Two_Streams);
  printf("Enhancement_Layer_Bitstream_Filename = %s\n", Enhancement_Layer_Bitstream_Filename);
  printf("Big_Picture_Flag                     = %d\n", Big_Picture_Flag);
  printf("Main_Bitstream_Flag                  = %d\n", Main_Bitstream_Flag);
  printf("Main_Bitstream_Filename              = %s\n", Main_Bitstream_Filename);
  printf("Verify_Flag                          = %d\n", Verify_Flag);
  printf("Stats_Flag                           = %d\n", Stats_Flag);
  printf("User_Data_Flag                       = %d\n", User_Data_Flag);

}
#endif


//
// Check CPU type for MMX extensions
//
// code from John Schlichter's AVI2MPG1 v1.7
//
static 
void test_cpu( BOOL *p_fMMX, BOOL *p_f3DNOW )
{
	BOOL m_fMMX = FALSE;
	BOOL m_f3DNOW = FALSE;

	if ( p_fMMX == NULL || p_f3DNOW == NULL )
	{
		fprintf( stderr, "\nTestCPU() : p_fMMX or p_f3DNOW == NULL!");
		exit( ERROR );
	}

	printf(" ");
	// based on code from www.amd.com 3dNow manual 21928e.pdf
	//
	// Added define because compiler did not like CPUID instruction
	//
//	#define cpuid __asm _emit 0x0F __asm _emit 0xA2
    __asm
    {
		pushfd                          ; // save EFLAGS
		pop   eax                       ; // store EFLAGS in EAX
		mov   ebx, eax                  ; // save EBX for later testing
		xor   eax, 0x00200000           ; // toggle bit 21
		push  eax                       ; // put to stack
		popfd                           ; // save changed EAX to EFLAGS
		pushfd                          ; // push EFLAGS to TOS
		pop   eax                       ; // store EFLAGS in EAX
		cmp   eax, ebx                  ; // see if bit 21 has changed
		jz    NO_CPUID                  ; // if no change, no CPUID

		mov   eax, 0x80000000           ; // query for AMD extended functions
        cpuid                           ; // get extended function limit
		cmp   eax, 0x80000000           ; // is 8000_0001h supported?
		jbe   NO_EXTENDEDMSR            ; // if not, 3DNow! tech. not supported

		mov   [m_fMMX], 1                ; // but 3DNow is!
		mov   eax, 0x80000001           ; // setup extended function 1
		cpuid                           ; // call the function
		test  edx, 0x80000000           ; // test bit 31
		jz    NO_3DNOW                  ; // 3DNow Not Detected
        mov   [m_f3DNOW], 1              ; // 3DNow detected!
		jmp   GOT_3DNow

NO_EXTENDEDMSR:
		mov   eax, 0x00000001           ; // query for standard INTEL
        cpuid   						; // cpu feature flags
		test  edx, 0x00800000           ; // is MMX supported?
		jz    NO_MMX                    ;
		mov   [m_fMMX], 1                ;
GOT_3DNow:
NO_MMX:
NO_3DNOW:
NO_CPUID:
	//	pop eax;
	}	// end __asm

	printf(" ");

	// reset flags to false
	(*p_fMMX) = m_fMMX;
	(*p_f3DNOW) = m_f3DNOW;
}


//initializes the g_scan[][] matrix
// type = 0   normal scan-order (non transpose)
// type = 1   transpose scan-order (for transposing iDCT functions)
static void
init_scan_matrix( int type )
{
	int i;

	for ( i = 0; i < 64; ++i )
	{
		if ( type != SCAN_ORDER_NORMAL )
		{	// transpose scan order
			g_scan[0][i] = scan_tr[0][i];
			g_scan[1][i] = scan_tr[1][i];
		}
		else
		{	// normal scan order (non-transpose)
			g_scan[0][i] = scan_no[0][i];
			g_scan[1][i] = scan_no[1][i];
		}
	}
}