/* store.cpp, picture output routines                                         */

/* Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The MPEG Software Simulation Group does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 *
 * Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
 * are subject to royalty fees to patent holders.  Many of these patents are
 * general enough such that they are unavoidable regardless of implementation
 * design.
 *
 *
 * MPEG2AVI
 * --------
 * v0.16B35 corrected memset() bug in store_avi_gdi() which didn't
 *    clear the entire framebuffer workspaces
 *    increased memory padding on all calloc() calls
 * v0.16B34 converted malloc() calls to calloc() calls
 * v0.16B33 for '-*', a square-block is printed for each I-frame
 * v0.16B32 for '-*', a correct upper-left justification is applied so the
 *    the mpeg-header info is always in the UL corner, regardless of cropping
 * v0.16B30 update_mpeginfo() called in Write_Frame(), this updates the LBA,
 *          temporal_reference, pictref, etc. variables
 * 
 * v0.16B23 if (UserOptIDCT == _IDCT_MMX_AAN), disallow AVI-writing!
 *         moved the planar->packed conversion to yuv2yuv.c.  mildly changed
 *          the calling parameters (now uses a ConvBlock struct)
 *         in store_avi_gdi(), fb_yuv is malloc'd to a 32bpp space (instead of 24bpp)
 * 
 * v0.16B22 __asm emms to the end of Write_Frame() & middle of store_avi_gdi()
 *         MMX optimizations to YV12, YUY2 buffer generation.
 *         added EXTRA_MEM_PAD to extend frame-buffer stores.
 * v0.16B2 added watermarking feature to frames,
 *           1st Decoded frame counter (always starts from 0)
 *           2nd line gop_timecode (from GOP header)
 *           3rd line temporal reference (from MPEG picture header)
 *           4th line VOB LBA address of last good MPEG SYSTEM HEADER
 * v0.16B1 ++g_cbMPEGpicture moved to Write_Frame(), so it doesn't double-count
 *       interlaced (two fields) video
 *       moved keyboard parsing to here
 *       added m_cbSkipFrame to store_avi_one() to implement UserOptAVISkipFrame
 *        the skip-feature applies to GDI/AVI output, but it occurs after the
 *        frame is sent through the IVTC (so as to not disturb its operatino)
 * v0.15B51 the UserOptFRangeEnd triggers -> g_cmd_buttonStop 
 * v0.15B5 restructured post-processing operations, fixed malloc() bug in
 *       store_avi_gdi()
 *
 *       | MPEG2V12 | -> | VRESIZE.C | -> | YUV2YUV | -> | CROP&CSC | -> OUTPUT
 *       MPEG decoder    -3X, -3Y resize   pixFormat     crop viewport
 *                                         conversion
 *
 * v0.15B4 direct-YUV AVI mode.  The decoded-MPEG frames are transferred
 *       to storeavi.c in the native YCbCr420 color-format
 *       half-resolution mode now works for all CHROMA formats (420/422/444)
 * v0.15 added frame-counter, g_FrameCount.  This counts the #frames received
 *       from the MPEG-decoder, and is not 100% reliable.  It's a useful
 *       statistic for the GDI and AVI functions, though.
 *       
 *       added process_filter_cmd( void ) to proces g_cmd_filter
 *
 * v0.14
 *    Clear all memory-spaces, 
 *    Created a new function store_avi_gdi() ... 
 *    half-resolution output mode (currently only supports CHROMA420 foramt.)
 *    The 1/2-res mode only supports YUV420 planar, MPEG's most popular
 *    color format.
 *
 *     Cropping & 1/2-resolution implementation
 *     ----------------------------------------
 *
 *     1)  decoder-output <src[0],src[1],src[2]>
 *     2)  Half-resolution?  if yes, downsample
 *     3)  build YUV888 24bpp packed bitmap from <Resized image, YUV444 planar)
 *         (crop & convert image simultaneously)
 *     4)  call output functions (pass YUV888 24bpp packed image)
 *     5)  (display_dib or store_avi_one), convert color space with yuv2rgb24() 
 *         yuv2rgb24()...
 *     6)  done
 *
 *     only store_avi_gdi() implement cropping/downsampling...the other
 *     functions ignore these operations.
 *
 *    When 1/2-res mode is in use, store_avi_gdi() *directly* calls 
 *    ivtc_half_8bpp() [ telecine.c ], then sends the processed frame to
 *    the appropriate output-routine ( store_avi() or display_dib() )
 *
 *    By handling the telecine-conversion in native YUV420 planar, the
 *    upscaling/downscaling (YUV420->422->444) filters are bypassed entirely,
 *    increasing both performance and precision.
 *
 *    Overall data-flow from MPEG-decoder's output to file
 *
 *    1)  store_one()  -> src[0],src[1],src[2] -> store_avi()
 *    2)  src[0]... -> src[0] (y_half), src[1] ->u444, src[2] -> v444 )
 *    3)     fb_yuv -> ivtc_half_8bpp() -> (display_dib or store_avi_one) -> yuv2rgb24()
 *
 * v0.11
 * Most modifications are to store_ppm_tga() ... in addition to its previous
 * PPM/TGA output, this function now handles GDI and AVI.
 *
 * store_ppm_tga() redirects GDI output to display_dib() [windisp.c]
 *					 and AVI to store_avi_one() [storeavi.c]
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <io.h>
#include <conio.h>

#include "config.h"
#include "global.h"
#include <windows.h> /* used only to get type declaration BOOL */
#include <wingdi.h> /* used for BITMAPINFO */

void yuv444torgb32( tImageOpBlock *pConvBlock );


/* private prototypes */
static void store_one _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height));
//static void store_yuv _ANSI_ARGS_((char *outname, unsigned char *src[],
//  int offset, int incr, int height));
//static void store_sif _ANSI_ARGS_((char *outname, unsigned char *src[],
//  int offset, int incr, int height));
//static void store_ppm_tga _ANSI_ARGS_((char *outname, unsigned char *src[],
//  int offset, int incr, int height, int tgaflag));
static void store_avi_gdi _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height, int tgaflag));
//static void store_yuv1 _ANSI_ARGS_((char *name, unsigned char *src,
//  int offset, int incr, int width, int height));
//static void putbyte _ANSI_ARGS_((int c));
//static void putword _ANSI_ARGS_((int w));
BOOL process_filter_cmd( void ); /* process command-variable g_cmd_filter */

	/* private variables */

static RECT srcRect; /* resolution of MPEG-video */
static BITMAPINFO g_bmi;
#define OBFRSIZE 4096
static unsigned char obfr[OBFRSIZE];
static unsigned char *optr;
static int outfile;

#define EXTRA_MEM_PAD 16	// extra bytes to pad each framebuffer
	// 0.16B22  EXTRA_MEM_PAD allows MMX-operations to read beyond the
	// logical end of a video-frame (for example, an MMX-loop might
	// right up to 7 bytes beyond the end of the frame, for DWORD/QWORD
	// alignment purposes.)


int
rstrcmpi( const char a[], const char b[] )
{
	long counter = 0;
	long result = 0;
	long len_a = 0, len_b = 0;
	long minab = 0;
	
	if ( a != NULL )
		len_a = strlen( a );

	if ( b != NULL )
		len_b = strlen( b );

	if ( len_a == 0 && len_b == 0 )
		return 0;	/* equal ?!? */

	if ( len_a == 0 )
		return -1; /* left < right */

	if ( len_b == 0 )
		return 1; /* left > right */

	minab = ( len_a > len_b ) ? len_b : len_a;

	for ( counter = 0; counter < minab && result == 0; ++counter )
	{
		if ( toupper( a[ len_a - counter ] ) !=
			toupper( b[ len_b - counter ] ) )
			result = toupper( a[ len_a - counter ] ) >
				toupper( b[ len_b - counter ] ) ? 1 : -1 ;
	}

	return result;
}

/*
 * store a picture as either one frame or two fields
 */
void Write_Frame(src,frame)
unsigned char *src[];
int frame;
{
  char outname[FILENAME_LENGTH];

  update_mpeginfo();	// update LBA, goptimecode, temporalref, v0.16B3

  if (progressive_sequence || progressive_frame || Frame_Store_Flag)
  {
    /* progressive */
    sprintf(outname,Output_Picture_Filename,frame,'f');
    store_one(outname,src,0,Coded_Picture_Width,vertical_size);
  }
  else
  {
    /* interlaced */
    sprintf(outname,Output_Picture_Filename,frame,'a');
    store_one(outname,src,0,Coded_Picture_Width<<1,vertical_size>>1);

    sprintf(outname,Output_Picture_Filename,frame,'b');
    store_one(outname,src,
      Coded_Picture_Width,Coded_Picture_Width<<1,vertical_size>>1);
  }
  ++g_cbMPEGPicture; /* update decoded picture counter */
  //g_cbMPEGPicture = frame;

 // __asm emms;
}

/*
 * store one frame or one field
 */
static void store_one(outname,src,offset,incr,height)
char *outname;
unsigned char *src[];
int offset, incr, height;
{
  static int first_iteration = 1;
  int keyp; // keypress
  if ( first_iteration )
  {
	  __asm emms;

	  first_iteration = 0;
	  //g_FrameCount = 0;  /* initialize frame counter */
	  printf("\nMPEG-%d video resolution : <%d x %d>, coded format = ",
		  (1+base.MPEG2_Flag), Coded_Picture_Width,Coded_Picture_Height );
	  switch( chroma_format )
	  {
	    case CHROMA420 :
		  printf("YUV420");
		  break;
	    case CHROMA422 :
		  printf("YUV422");
		  break;
	    case CHROMA444 :
		  printf("YUV444");
		  break;
	    default:
		  printf("unknown!");
	  } /* endswitch */

	  // px3 ver 0.1.0, autoresizer first version
	  if ( GETOPT(UserOptAutoResize) > 0 && GETOPT(UserOptAutoResize) < 9 )
	  {
		  	// first reset all -X and -1 Params
			CLEAROPT( UserOptResizeX );
			CLEAROPT( UserOptResizeY );

			// now we calc the resize based on the mpeg-stream data
			// for better preview we reduce y
			// horizontal_size  vertical_size
			switch(aspect_ratio_information)
			{
				case 2: // 4:3
					if ( GETOPT(UserOptAutoResize) == 1 )
						SETOPT( UserOptResizeY, (int)(horizontal_size/4)*3 )
					else
					{
						switch( GETOPT(UserOptAutoResize) )
						{
							case 2:
								SETOPT( UserOptResizeY, (int)(horizontal_size/1.25) );
								break;
							case 3:
								SETOPT( UserOptResizeY, (int)(horizontal_size/1.33) );
								break;
							case 4:
								SETOPT( UserOptResizeY, (int)(horizontal_size/1.36) );
								break;
						}
					}
					break;
				case 3: // 16:9
					if ( GETOPT(UserOptAutoResize) == 1 )
						SETOPT( UserOptResizeY, (int)(horizontal_size/16)*9 )
					else
					{
						switch( GETOPT(UserOptAutoResize) )
						{
							case 5:
								SETOPT( UserOptResizeY, (int)(horizontal_size/1.66) );
								break;
							case 6:
								SETOPT( UserOptResizeY, (int)(horizontal_size/1.77) );
								break;
							case 7:
								SETOPT( UserOptResizeY, (int)(horizontal_size/1.85) );
								break;
							case 8:
								SETOPT( UserOptResizeY, (int)(horizontal_size/2.25) );
								break;
						}
					}
					break;
			}

			// check for div 8 (mmx scale)
			if ( GETOPT( UserOptResizeY )%8 )
			{
				SETOPT( UserOptResizeY, 
					(GETOPT( UserOptResizeY )-
					GETOPT( UserOptResizeY )%8));
			}

			switch(aspect_ratio_information)
			{
				case 2: // 4:3
					if ( GETOPT(UserOptAutoResize) == 1 )
						SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*4)/3 )
					else
					{
						switch( GETOPT(UserOptAutoResize) )
						{
							case 2:
								SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*1.25) );
								break;
							case 3:
								SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*1.33) );
								break;
							case 4:
								SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*1.36) );
								break;
						}
					}
					break;
				case 3: // 16:9
					if ( GETOPT(UserOptAutoResize) == 1 )
						SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*16)/9 )
					else
					{
						switch( GETOPT(UserOptAutoResize) )
						{
							case 5:
								SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*1.66) );
								break;
							case 6:
								SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*1.77) );
								break;
							case 7:
								SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*1.85) );
								break;
							case 8:
								SETOPT( UserOptResizeX, (int)(GETOPT( UserOptResizeY )*2.25) );
								break;
						}
					}

					break;
			}

			// check for div 8 (mmx scale)
			if ( GETOPT( UserOptResizeX )%8 )
			{
				SETOPT( UserOptResizeX, 
					(GETOPT( UserOptResizeX )-
					GETOPT( UserOptResizeX )%8));
			}


			if ( GETOPT(UserOptAutoResize) > 1 )
			{
				printf("\n\nAutoresizer sets values to X=%d, Y=%d (", GETOPT( UserOptResizeX ), GETOPT( UserOptResizeY ) );
				switch( GETOPT(UserOptAutoResize) )
				{
					case 2:
						printf("~1.25:1)");
						break;
					case 3:
						printf("~1.33:1)");
						break;
					case 4:
						printf("~1.36:1)");
						break;
					case 5:
						printf("~1.66:1)");
						break;
					case 6:
						printf("~1.77:1)");
						break;
					case 7:
						printf("~1:85:1)");
						break;
					case 8:
						printf("~2.35:1)");
						break;
				}
			}
			else
				printf("\n\nAutoresizer sets values to X=%d, Y=%d", GETOPT( UserOptResizeX ), GETOPT( UserOptResizeY ) );
	  } // of autoresizer

	  // Apect ratio px3 ver 0.1.0
		printf("\nAspect ratio from MPEGstream: ");
		switch(aspect_ratio_information)
		{
			case 1:
				printf( "[1] square pixels" );
				break;
			case 2:
				printf( "[2] 4:3 display" );
				printf( "\n / ratio = %.2f:1\nCurrent settings produce an ratio of %.2f:1\nRecommended: 1.25:1 / 1.33:1 / 1.36:1", 
					(double)horizontal_size/vertical_size,
					(double)GETOPT(UserOptResizeX)/GETOPT(UserOptResizeY));
				break;
			case 3:
				printf( "[3] 16:9 display" );
				printf( " / ratio = %.2f:1\nCurrent settings produce an ratio of %.2f:1\nRecommended: 1.66:1 (Europe) / 1.77:1 (GB) / 1:85:1 (US) / 2.35:1 (CinemaScope)\n", 
					(double)horizontal_size/vertical_size,
					(double)GETOPT(UserOptResizeX)/GETOPT(UserOptResizeY));
				break;
			case 4:
				printf( "[4] 2.21:1 display" );
				break;
			default:
				printf( "[?] unknown aspect-ratio" );
				break;
		} // of aspect ratio


	  /* setup rectangle for store.c */
	  srcRect.left = 0; 
	  srcRect.top = 0;
	  srcRect.right = Coded_Picture_Width;
	  srcRect.bottom = Coded_Picture_Height;
	  printf("\n");
  } /* endif ( first_iteration ) */
  
  switch (Output_Type)
  {
  case T_YUV:
//    store_yuv(outname,src,offset,incr,height);
    break;
  case T_SIF:
//    store_sif(outname,src,offset,incr,height);
    break;
  case T_TGA:
//    store_ppm_tga(outname,src,offset,incr,height,1);
    break;
  case T_PPM:
//    store_ppm_tga(outname,src,offset,incr,height,0);
    break;
#ifdef DISPLAY
  case T_X11:
    dither(src);
    break;
#endif
  case T_GDI:
	store_avi_gdi(outname,src,offset,incr,height,T_GDI);
	break;
  case T_AVI:
	store_avi_gdi(outname,src,offset,incr,height,T_AVI);
	break;
  case T_AVI_YV12:  // quick hack for 0.15B4
	store_avi_gdi(outname,src,offset,incr,height,T_AVI_YV12);
	break;
  case T_AVI_YUY2:  // quick hack for 0.15B5
	store_avi_gdi(outname,src,offset,incr,height,T_AVI_YUY2);
	break;
  default:
	  fprintf(stderr, "\nstore_one() : Unknown output mode -o%d", Output_Type );
	  exit(ERROR);
    break;
  }

  if ( Output_Type != T_GDI )
  { // scan keyboard
  	if ( _kbhit() )	// recent keypress?  If not, don't _getch()
	{
		keyp = _getch();	// yes, read keypress
		if ( toupper( keyp ) == toupper( KEY_QUIT ) )
		{
			g_cmd_buttonStop = TRUE;	// set command to stop
		}
	}

	if ( g_cmd_buttonStop )
	{
		printf("\nDecoder stopped.");
		avi_fileclose();
		exit( ERROR );
	}
  }
}

/* separate headerless files for y, u and v */
/*
static void store_yuv(outname,src,offset,incr,height)
char *outname;
unsigned char *src[];
int offset,incr,height;
{
  int hsize;
  char tmpname[FILENAME_LENGTH];

  hsize = horizontal_size;

  sprintf(tmpname,"%s.Y",outname);
  store_yuv1(tmpname,src[0],offset,incr,hsize,height);

  if (chroma_format!=CHROMA444)
  {
    offset>>=1; incr>>=1; hsize>>=1;
  }

  if (chroma_format==CHROMA420)
  {
    height>>=1;
  }

  sprintf(tmpname,"%s.U",outname);
  store_yuv1(tmpname,src[1],offset,incr,hsize,height);

  sprintf(tmpname,"%s.V",outname);
  store_yuv1(tmpname,src[2],offset,incr,hsize,height);
}
*/
/* auxiliary routine 
static void store_yuv1(name,src,offset,incr,width,height)
char *name;
unsigned char *src;
int offset,incr,width,height;
{
  int i, j;
  unsigned char *p;

  if (!Quiet_Flag)
    fprintf(stderr,"saving %s\n",name);

  if ((outfile = _open(name,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(Error_Text,"Couldn't create %s\n",name);
    Error(Error_Text);
  }

  optr=obfr;

  for (i=0; i<height; i++)
  {
    p = src + offset + incr*i;
    for (j=0; j<width; j++)
      putbyte(*p++);
  }

  if (optr!=obfr)
    _write(outfile,obfr,optr-obfr);

  _close(outfile);
}*/

/*
 * store as headerless file in U,Y,V,Y format
 *
static void store_sif (outname,src,offset,incr,height)
char *outname;
unsigned char *src[];
int offset, incr, height;
{
  int i,j;
  unsigned char *py, *pu, *pv;
  static unsigned char *u422, *v422;

  if (chroma_format==CHROMA444)
    Error("4:4:4 not supported for SIF format");

  if (chroma_format==CHROMA422)
  {
    u422 = src[1];
    v422 = src[2];
  }
  else
  {
    if (!u422)
    {
      if (!(u422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                           *Coded_Picture_Height)))
        Error("malloc failed");
      if (!(v422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                           *Coded_Picture_Height)))
        Error("malloc failed");
    }
  
    conv420to422(src[1],src[2],u422,v422, Coded_Picture_Width,
		Coded_Picture_Height );
  }

  strcat(outname,".SIF");

  if (!Quiet_Flag)
    fprintf(stderr,"saving %s\n",outname);

  if ((outfile = _open(outname,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(Error_Text,"Couldn't create %s\n",outname);
    Error(Error_Text);
  }

  optr = obfr;

  for (i=0; i<height; i++)
  {
    py = src[0] + offset + incr*i;
    pu = u422 + (offset>>1) + (incr>>1)*i;
    pv = v422 + (offset>>1) + (incr>>1)*i;

    for (j=0; j<horizontal_size; j+=2)
    {
      putbyte(*pu++);
      putbyte(*py++);
      putbyte(*pv++);
      putbyte(*py++);
    }
  }

  if (optr!=obfr)
    _write(outfile,obfr,optr-obfr);

  _close(outfile);
}*/

/*
 * store as PPM (PBMPLUS) or uncompressed Truevision TGA ('Targa') file
 *
static void store_ppm_tga(outname,src,offset,incr,height,tgaflag)
char *outname;
unsigned char *src[];
int offset, incr, height;
int tgaflag;
{
  int i, j;
  int y, u, v, r, g, b;
  int crv, cbu, cgu, cgv;
  unsigned char *py, *pu, *pv;
  static unsigned char tga24[14] = {0,0,2,0,0,0,0, 0,0,0,0,0,24,32};
  char header[FILENAME_LENGTH];
  unsigned char *pixelxy_addr = NULL;
  static unsigned char *u422, *v422, *u444, *v444;
  
  
  if (chroma_format==CHROMA444)
  {
    u444 = src[1];
    v444 = src[2];
  }
  else
  {
    if (!u444)
    {
      if (chroma_format==CHROMA420)
      {
        if (!(u422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                             *Coded_Picture_Height)))
          Error("malloc failed");
        if (!(v422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                             *Coded_Picture_Height)))
          Error("malloc failed");
      }

      if (!(u444 = (unsigned char *)malloc(Coded_Picture_Width
                                           *Coded_Picture_Height)))
        Error("malloc failed");

      if (!(v444 = (unsigned char *)malloc(Coded_Picture_Width
                                           *Coded_Picture_Height)))
        Error("malloc failed");
    }

    if (chroma_format==CHROMA420)
    {
      conv420to422(src[1],src[2],u422,v422, Coded_Picture_Width, 
		  Coded_Picture_Height );
      conv422to444(u422,v422,u444,v444,Coded_Picture_Width, 
		  Coded_Picture_Height );
    }
    else
    {
      conv422to444(src[1],src[2],u444,v444,Coded_Picture_Width, 
		  Coded_Picture_Height );
    }
  }

  //strcat(outname,tgaflag ? ".tga" : ".ppm");
  switch ( tgaflag )
  {
	case 0 : if ( rstrcmpi( outname, ".ppm" ) != 0 )
				 strcat( outname, ".ppm" );
		break;
	case 1 : if ( rstrcmpi( outname, ".tga" ) != 0 )
				 strcat( outname, ".tga" );
		break;
	default :
		;
  }
  

  if ( tgaflag )
  {
  if (!Quiet_Flag)
    fprintf(stderr,"saving %s\n",outname);

  if ((outfile = _open(outname,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(Error_Text,"Couldn't create %s\n",outname);
    Error(Error_Text);
  }
  }

  optr = obfr;

  if (tgaflag)
  {
    // TGA header 
    for (i=0; i<12; i++)
      putbyte(tga24[i]);

    putword(horizontal_size); putword(height);
    putbyte(tga24[12]); putbyte(tga24[13]);
  }
  else
  {
    // PPM header 
    sprintf(header,"P6\n%d %d\n255\n",horizontal_size,height);

    for (i=0; header[i]!=0; i++)
      putbyte(header[i]);
  }

  // matrix coefficients 
  crv = Inverse_Table_6_9[matrix_coefficients][0];
  cbu = Inverse_Table_6_9[matrix_coefficients][1];
  cgu = Inverse_Table_6_9[matrix_coefficients][2];
  cgv = Inverse_Table_6_9[matrix_coefficients][3];
  
  for (i=0; i<height; i++)
  {
    py = src[0] + offset + incr*i;
    pu = u444 + offset + incr*i;
    pv = v444 + offset + incr*i;

    for (j=0; j<horizontal_size; j++)
    {
      u = *pu++ - 128;
      v = *pv++ - 128;
      y = 76309 * (*py++ - 16); // (255/219)*65536 
      r = Clip[(y + crv*v + 32768)>>16];
      g = Clip[(y - cgu*u - cgv*v + 32768)>>16];
      b = Clip[(y + cbu*u + 32786)>>16];
	  
	  if (tgaflag)
	  {
	        putbyte(b); putbyte(g); putbyte(r);
	  }
	  else
	  {
	        putbyte(r); putbyte(g); putbyte(b);
	  }
    }
  }

  if (optr!=obfr)
  {
    _write(outfile,obfr,optr-obfr); // write image to file 
    _close(outfile);
  }
}*/

/*
 * store as AVI (Windows) or display to Windows GDI
 */
static void store_avi_gdi(outname,src,offset,incr,height,tgaflag)
char *outname;
unsigned char *src[];
int offset, incr, height;
int tgaflag;
{
  char szText[256], szName[256]; // avi description string
  /*int y, u, v, r, g, b;*/
  DWORD y_src_addr=0, uv_src_addr=0, dest_addr=0;
  RECT dstRect;
  BITMAPINFOHEADER bmih;
  DWORD dwPixFormat = 0;	/* needed for avi_fileopen() */
  DWORD dwXpos, dwYpos; /* v0.16B32 justification (in #chars) for vtitler */
  DWORD psize;

  static unsigned char *u422 = NULL, *v422= NULL, *u444=NULL, *v444=NULL,
	  *y_full=NULL, *u_extra = NULL, *v_extra = NULL, *y_extra = NULL;
  
  static int first_iteration = 1;
  static unsigned char *fb_yuv = NULL, *y_ptr = NULL , *u_ptr = NULL,
	  *v_ptr = NULL;
  static long m_OutputHeight, m_OutputWidth; 
	/* dimensions of DISPLAYED (final) frames, after cropping */
  static long m_cbSkipFrame = 0;
    /* v0.16B1 AVISkipFrame counter */

  static RECT CropRect; /* cropping rectangle, passed to -o6 and -o7 */

  /* the cropping rectangle is with respect to Resized_Picture_xxx
     yuv2rgb24 ultimately handles the cropping. */
  tImageOpBlock convBlock; /* conversion information for yuv2yuv() */

  if ( first_iteration )
  {

	first_iteration = 0;	/* disable this flag */
	psize = (Coded_Picture_Width * Coded_Picture_Height) >> 5;
  	/* allocate memory and clear it */
	fb_yuv = (unsigned char *)calloc( (psize*4)+2, 32 ); 

	y_full = (unsigned char *)calloc(  psize + 2, 32 );
	/* y, u and v extra are for the resizer */
	y_extra = (unsigned char *)calloc( psize + 2, 32);
	u_extra = (unsigned char *)calloc( psize + 2, 32);
	v_extra = (unsigned char *)calloc( psize + 2, 32);
	u422 = (unsigned char *)calloc( psize + 2, 32);
	v422 = (unsigned char *)calloc( psize + 2, 32);
	u444 = (unsigned char *)calloc( psize + 2, 32);
	v444 = (unsigned char *)calloc( psize + 2, 32 );


	// px3 ver 0.1.0, autoresizer first step
/*	if ( GETOPT(UserOptAutoResize) == 1 )
	{
		// first rest all -X and -1 Params
		CLEAROPT( UserOptResizeX );
		CLEAROPT( UserOptResizeY );
		CLEAROPT( UserOptCropX );
		CLEAROPT( UserOptCropY );

		// now we calc the resize based on the mpeg-stream data
		// for better preview we reduce y
//					(double)horizontal_size/vertical_size,
		switch(aspect_ratio_information)
		{
			case 2: // 4:3
				SETOPT( UserOptResizeY, (int)(vertical_size/4)*3 );
				break;
			case 3: // 16:9
				SETOPT( UserOptResizeY, (int)(vertical_size/16)*9 );
				break;
		}


		SETOPT( UserOptResizeX, horizontal_size )

		printf("X = %d , Y=%d\n", GETOPT( UserOptResizeX ), GETOPT( UserOptResizeY ) );
	}
*/
	/* vresizer.c */
	if ( CHECKOPT( UserOptResizeX ) && UserOptResizeX.data < 32 
		 || CHECKOPT( UserOptResizeY ) && UserOptResizeY.data < 32 )
	{
		fprintf( stderr, "\nshrink_frame() : minimum size requirement (32x32) not met!");
		exit( ERROR );
	}


	if ( !( fb_yuv && y_full && y_extra && u_extra && v_extra &&
		u422 && v422 && u444 && v444 ) )
	{
		fprintf( stderr, 
			"\nstore_avi_gdi() : could not allocate temp framebuffers!");
		exit( ERROR );
	}
	else
	{
		/* clear the framebuffers! */
		psize = ( psize +2 ) << 5;	// convert back to #bytes
		memset((LPVOID)fb_yuv,0,psize*4);
		memset((LPVOID)y_full,0,psize );
		memset((LPVOID)y_extra,0,psize );
		memset((LPVOID)u_extra,128,psize );
		memset((LPVOID)v_extra,128,psize );
		memset((LPVOID)u422,128,psize );
		memset((LPVOID)v422,128,psize );
		memset((LPVOID)u444,128,psize );
	  	memset((LPVOID)v444,128,psize );
	}  /* endif */

	// v0.16B23, if (UserOptIDCT == _IDCT_MMX_AAN), disallow AVI-writing!
	if ( tgaflag != T_GDI )
	{
		if ( rstrcmpi( outname, ".avi" ) != 0 )
			strcat( outname, ".avi" );

		if ( GETOPT( UserOptIDCT ) == _IDCT_MMX_AAN )
		{
			fprintf( stderr, "\nCommandOption '-r%d' : You have chosen a fast (inaccurate) IDCT algorithm.",
				GETOPT( UserOptIDCT ) );
			fprintf( stderr, "\nAVI-writing is not permitted with low-precision IDCT.");
			fprintf( stderr, "\n(try '-r0', '-r1', or '-r2' instead!)\n");
			exit( ERROR );
		}
	}

	/* check chroma format if ( g_vHalfResFlag == TRUE )
	   currently we only support CHROMA420 for half-resolution mode */
	//if ( g_vHalfResFlag && chroma_format != CHROMA420 )
	//{
	//	fprintf( stderr, "\nstore.c - half-resolution requires color-format of input bitstream to be CHROMA420." );
	//	  Error("Color format incompatible");
	  //}

	/* load valid co-ordinates into Resized_Picture_Width/Height */
	// setup variables for DIB/AVI output routines 
	Resized_Picture_Width= g_vHalfResFlag ? (Coded_Picture_Width >> 1 ):
		Coded_Picture_Width;
	Resized_Picture_Height = g_vHalfResFlag ? (Coded_Picture_Height >> 1):
		Coded_Picture_Height;

	/* check if g_AVIcrop already has valid co-ordinates, if not set
	   them to current image-height/width...i.e. effectively disables cropping */
    /* m_OutputWidth, m_OutputHeight replace g_AVIcrop... */
	m_OutputWidth = Resized_Picture_Width;
	m_OutputHeight = Resized_Picture_Height;

	if ( CHECKOPT( UserOptCropX ) )
		m_OutputWidth = GETOPT( UserOptCropX );

	if ( CHECKOPT( UserOptCropY ) )
		m_OutputHeight = GETOPT( UserOptCropY );

	/* validate g_AVIcrop & Resized_Picture ... are parameters legal? */
	/* Make sure cropped sized is smaller than resized-image */
	if ( m_OutputWidth > Resized_Picture_Width || 
		m_OutputHeight > Resized_Picture_Height )
	{
		fprintf( stderr, "\nstore_avi_gdi() : Viewport dimensions (\"-1\" %ld X %ld)",
			m_OutputWidth , m_OutputHeight );
		fprintf( stderr, " exceed\n\tresized-output's dimensions (%ld X %ld) !",
			Resized_Picture_Width, Resized_Picture_Height );
		exit (ERROR );
	}

	/* v0.16B22, crop window must be divisible by 8 for MMX-crop functions! */

	if ( (m_OutputWidth & 0x07) || (m_OutputHeight & 0x03 ) )
	{
		fprintf( stderr, "\nstore_avi_gdi() : Viewport dimensions (\"-1\" %ld X %ld)",
			m_OutputWidth , m_OutputHeight );			
		fprintf( stderr, " are\n\tnot divisible by 8!" );
		exit( ERROR );
	}

	CropRect.top = ( Resized_Picture_Height - m_OutputHeight  ) >> 1;
	CropRect.bottom = Resized_Picture_Height - 
		(( Resized_Picture_Height - m_OutputHeight ) >> 1);
	CropRect.left = ( Resized_Picture_Width - m_OutputWidth ) >> 1;
	CropRect.right = Resized_Picture_Width -
		(( Resized_Picture_Width - m_OutputWidth ) >> 1 );


  	// create stream header text
    strcpy( szName, Version ); // Version = "mpeg2avi ..." 
	// v0.16B22, updated IDCT-type reporting
	switch( GETOPT( UserOptIDCT ) )
	{
		case _IDCT_MMX32 :
			sprintf( szText, " {IDCT32MMX} " ); // 32-bit MMX
			break;
		case _IDCT_FPU64 :
			sprintf( szText, " {IDCT64FPU} " ); // 64-bit double float
			break;		
		case _IDCT_MMX_CHEN1 :
			sprintf( szText, " {IDCT16MMXChen} " ); // 16-bit MMX Chen
			break;
		case _IDCT_MMX_AAN :	// this output mode is no longer allowed
			sprintf( szText, " {IDCT16MMXAAN} " ); // 16-bit MMX AAN
			break;
		default :
			sprintf( szText, " {IDCT ???} " ); // unknown IDCT
	}
	strcat( szName, szText );
	sprintf( szText, "%s", g_fastfilter ? "'-q1'" : "'-q0'" );
	strcat( szName, szText );
	if ( CHECKOPT( UserOptIVTC ) ) // is IVTC in use?
	{
		sprintf( szText, ", -s%d IVTC(%d)", GETOPT( UserOptIVTC ), 
			GETOPTAUX( UserOptIVTC )  ); // extended IVTC config info 
		strcat( szName, szText );
	}

	if ( tgaflag != T_GDI )
	{
		// construct bitmapinfo
		bmih.biSize = sizeof(BITMAPINFOHEADER); 
		bmih.biWidth = m_OutputWidth;
		bmih.biHeight = m_OutputHeight;
		// some codecs *require* valid Height/Width params to avoid ICERR_BADFORMAT
		bmih.biPlanes = 1;			// must be 1
		//m_bmih.biSizeImage = 0;			// okay to be 0 for BI_RGB
		bmih.biXPelsPerMeter = bmih.biYPelsPerMeter = 0;
		bmih.biClrUsed = bmih.biClrImportant = 0;
		bmih.biSizeImage = bmih.biWidth * bmih.biHeight; 

		switch ( Output_Type )
		{
		case T_AVI_YV12 :
//			fUseYUV = TRUE;
			dwPixFormat = mmioFOURCC( 'Y','V','1','2');
			bmih.biBitCount = 12;
			bmih.biSizeImage = ( bmih.biSizeImage * 3 ) >> 1;
			break;
		case T_AVI_YUY2 :
//			fUseYUV = TRUE;
			dwPixFormat = mmioFOURCC( 'Y','U','Y','2');
			bmih.biBitCount = 16;
			bmih.biSizeImage = ( bmih.biSizeImage ) << 1;
			break;
		default :
			dwPixFormat = BI_RGB;
			bmih.biBitCount = 24;
			bmih.biSizeImage = ( bmih.biSizeImage * 3 );
//			fUseYUV = FALSE;
		} /* endswitch ( Output_Type ) */

		bmih.biCompression = dwPixFormat;	// standard RGB bitmap
		g_bmi.bmiHeader = bmih;

		__asm emms;	

		avi_fileopen( outname, szName, &CropRect, g_bmi, UserOptAVIrate, 
			UserOptAVISplitLength, TRUE );
		// the TRUE requests a codec-selection dialog-box

	}

	m_cbSkipFrame = 0;

  } /* endif ---- if ( first_iteration ) */


  // Added if by PX3 .. no need for keycheck if we're in
  // AVI-Output.. 
  if ( tgaflag == T_GDI )
  {
	process_filter_cmd( );  /* process commands to filter routines */
	process_resize_cmd( );  /* process commands to resize routines */
  }
	
  /* check UserOptFRangeStart & UserOptFRangeEnd */

  /* gethdr.c skips B-frames, but we still need to skip I/P frame processing
     we can't simply skip I/P frames entirely, because we might start
	 displaying on a B-frame which depends on previous I/P frames! */
  /* the inequalities imply that <a> ... <b> are INCLUDED in the range! */
  if ( CHECKOPT( UserOptFRangeStart ) )
  {
	  if ( GETOPT( UserOptFRangeStart ) >= g_cbMPEGPicture)
	  {
		  return;
	  }

	  //PX3 moved here ver 0.1.0
	  if ( CHECKOPT( UserOptFRangeEnd ) )
	  {
		 if ( g_cbMPEGPicture > GETOPT( UserOptFRangeEnd ) )
		 {
			 printf("\r< FRangeEnd > = MPEG-FrameCount (%ld)>", g_cbMPEGPicture );
			 g_cmd_buttonStop = TRUE;  /* stop after current frame */
		 }
	  }
  }


  y_ptr = src[0];
  u_ptr = src[1];
  v_ptr = src[2];

//  if ( !g_vHalfResFlag ) /* no 1/2 downsampling, FULL resolution */
//  {
	//y_ptr = ivtc_full_8bpp( src[ 0 ], g_FrameCount );
  y_ptr = ivtc_full_8bpp( src[ 0 ], g_cbMPEGPicture );
	
	/* Resize the output image */
	if ( y_ptr != NULL )  /* did ivtc delete current frame? */
	{ /* current frame wasn't deleted, resize the output! */
		if ( CHECKOPT( UserOptResizeX ) || CHECKOPT( UserOptResizeY ) )
		{
			dstRect.top = 0;
			dstRect.left = 0;
			dstRect.right = CHECKOPT( UserOptResizeX ) ? UserOptResizeX.data :
				srcRect.right;
			dstRect.bottom = CHECKOPT( UserOptResizeY ) ? UserOptResizeY.data :
				srcRect.bottom;

			shrink_frame( y_extra, u_extra, v_extra, y_ptr, u_ptr, v_ptr,
				&srcRect, &dstRect, chroma_format );

			  /* don't forget that RGB24 requires CHROMA444 to function */

			y_ptr = y_extra;
			u_ptr = u_extra;
			v_ptr = v_extra;
		}
	} /* endif ( y_ptr != NULL ) */

	/* at this point, the framebuffer y_ptr, u_ptr, v_ptr holds a resized
	   bitmap, whose format == chroma_format (specified by MPEG-bistream) */
  
  if ( y_ptr != NULL )  /* did ivtc delete current frame? */
  { /* current frame wasn't deleted, convert ouput format */

	/* Should we skip this frame? */
	if (  CHECKOPT( UserOptAVISkipEvery ) )
	{
	  if ( m_cbSkipFrame >= GETOPT( UserOptAVISkipEvery ) )
	  {  /* don't write this frame, skip it! */
		  m_cbSkipFrame = 0;
		  printf( "skipping! ('-!')                              " );
		  y_ptr = NULL;
		  return;
	  }
	  else
		  ++m_cbSkipFrame;	// don't skip current, so increment skip-counter
	}

	convBlock.srcRect = srcRect;
	// convBlock.destRect = not used
	convBlock.dwSrcFormat = chroma_format;
	switch ( tgaflag )
	{
		case T_AVI_YV12 :
			convBlock.dwDstFormat = CHROMA420;
			break;
		case T_AVI_YUY2 :
			convBlock.dwDstFormat = CHROMA422;
			break;
		default : /* RGb24 */
			convBlock.dwDstFormat = CHROMA444;
			break;
	} /* endswitch */
		
	/* specify input buffers */
	convBlock.src1 = y_ptr;
	convBlock.src2 = u_ptr;
	convBlock.src3 = v_ptr;
	/* supply allocated buffer space for operation output*/
	convBlock.dst1 = y_full;
	convBlock.dst2 = u444;
	convBlock.dst3 = v444;
		
	yuv2yuv( &convBlock );	// convert YUV(input) -> YUV(output)
	// input = MPEG-coded chroma format, always PLANAR
	// output= user selected output-mode, still PLANAR
	//
	// USER's CHOICE   | yuv2yuv() operation
	// -------------     -------------------
	//   "YV12"        =  output YUV420
	//   "YUY2"        =  output YUV422
	//   "RGB24"       =  output YUV444
		
	y_ptr = convBlock.outdst1; // get new y_ptr[]
	u_ptr = convBlock.outdst2; // get new u_ptr[]
	v_ptr = convBlock.outdst3; // get new v_ptr[]

	/* v0.16B2 print MPEG-frame code and GOP time codes, if requested */
	// note, under certain circumstances, the vtitler causes
	// visual 'side-effects.'  If y_ptr is a pointer to the *ORIGINAL* 
	// y_buffer, then vtitler() will actually modify the MPEG-decoder's
	// internal framebuffer state.  If the current-frame happens to be
	// an I/P frame, the watermarked-video is "carried forward" to
	// all dependent frames.  In essence, vtitler() can soil the video.
		
	// This only happens if the user plays the video without any
	// modification (resizing, 1/2 resolution)
	if ( CHECKOPT( UserOptPrintFrameCode ) )
	{
		// calculcate left-justification (in #chars) for titler
		dwXpos = ( CropRect.left >> 2 ) + 1; 
		dwYpos = ( CropRect.top >> 3 );

		// print decoded MPEG frame# on 1st line
		vt_printnum( (LPDWORD) y_ptr, g_cbMPEGPicture, dwXpos, dwYpos+1,
			( Resized_Picture_Width >>2 ) );

		// print GOP timecode on 2nd line, last char indicates "I" frame
		sprintf( szText, "%d:%02d:%02d:%02d%s%s",g_dtimecode.hour, 
			g_dtimecode.minute, g_dtimecode.second, g_dtimecode.frame,
			g_mpvpcur.pty == I_TYPE ? "I" : "",
			drop_flag ? "DROP":""  );
		vt_printstr( (LPDWORD) y_ptr, szText, dwXpos, dwYpos+2, 
			(Resized_Picture_Width>>2) );

		// print picture temporal reference on 3rd line
		vt_printnum( (LPDWORD) y_ptr, g_mpvpcur.tref, dwXpos, dwYpos+3, 
			(Resized_Picture_Width>>2) );

		// the one-frame delay mechanism accounts for the temporal
		// difference between the demuxer and MPEG-decoder states.
		if ( g_vobflag )
			vt_printnum( (LPDWORD) y_ptr, g_dwLBApack, dwXpos, dwYpos+4,
				( Resized_Picture_Width >>2 ) );
		//m_dwSystemPackLBAlast = g_dwSystemPackLBA;
	} // endif ( CHECKOPT( UserOptPrintFrameCode ) )

	/* prepare convBlock structure for yplanar2fb() function */

	// specify input buffers 
	convBlock.src1 = y_ptr;
	convBlock.src2 = u_ptr;
	convBlock.src3 = v_ptr;

	// supply allocated buffer space for operation output
	convBlock.dst1 = fb_yuv;
	convBlock.dst2 = NULL;
	convBlock.dst3 = NULL;

	convBlock.srcRect.left = CropRect.left; // specify top-left corner (src)
	convBlock.srcRect.top = CropRect.top;	// of viewport window

	convBlock.srcRect.right = Resized_Picture_Width; // src bitmap
	convBlock.srcRect.bottom = Resized_Picture_Height;	// dimensions

	convBlock.dstRect.right = m_OutputWidth; // specify viewport
	convBlock.dstRect.bottom = m_OutputHeight;	// dimensions

	/////////////////////////////////////////////////////////////////
	// Color-format conversion functions,
	// convert planar YUV to a single contiguous output-bitmap YUV
	//
	// planar :                  one_bitmap
	//   y_ptr[]                 
	//   u_ptr[] -> yuv2yuv.c -> fb_yuv[]
	//   v_ptr[]
	///////////////////////////////////////////////////////////////////

	switch ( tgaflag )
	{
	  case T_AVI_YV12 : /* convert YUV420 planar -> "YV12" planar 12bpp */
		yuv420toyv12( &convBlock );
		break;

	  case T_AVI_YUY2 : /* convert YUV422 planar -> "YUY2" packed 16bpp */
		yuv422toyuy2( &convBlock );
		break;

	  default: /* RGB24 */
		/* starting with version 0.16B22, the default case does not
		   perform any action.  yuv2rgb24() has been modified to
		   operate directly on the planar data y_ptr[], u_ptr[], v_ptr[].
		   This eliminates an wasteful repacking to 24bpp.  */

		/* (used to convert & crop YUV4:4:4 planar to YUV 24bpp packed pixel */
		/* note we do *NOT* perform color-space conversion just yet! */
		/* framebuffer[] = YUV 24bpp packed pixel image */
		/* invert image at the same time ) */
		 ;
		/* integer crop code no longer in use, has been combined with the
		   y444torgb24() function 
		for (j=0; j<m_OutputHeight ; j++)
		{
			for (i=0; i<m_OutputWidth ; i++)
			{
				y_src_addr = (j+CropRect.top)*Resized_Picture_Width +
					(i+CropRect.left);
				// the "-1" is necessary for proper alignment
				dest_addr = ( m_OutputHeight - j -1 ) * m_OutputWidth + i;
				*( fb_yuv + 3*dest_addr + 0 ) = *(y_ptr + y_src_addr );
				*( fb_yuv + 3*dest_addr + 1 ) = *(u_ptr + y_src_addr );
				*( fb_yuv + 3*dest_addr + 2 ) = *(v_ptr + y_src_addr );
			}
		}  // endfor 
		*/
	} // endswitch ( tgaflag )

	/////////////////////////////////////////////////////////////////
	// Final output stage
	// Perform 1 of 2 actions :
	//  1) display to WinGDI ( tgaflag == T_GDI)
	//
	//      or
	//
	//  2) write to AVI file (tgaflag != T_GDI)
	///////////////////////////////////////////////////////////////////

	if ( tgaflag == T_GDI )
	{
		yuv444torgb24( &convBlock ); // convert color y444 -> RGB24
		// display current frame to WinGDI
		display_dib( fb_yuv, &CropRect, g_cbMPEGPicture ); 
	}
	else if ( tgaflag == T_AVI || tgaflag == T_AVI_YV12 || tgaflag==T_AVI_YUY2 )
	{
		
		if ( tgaflag == T_AVI ) // RGB24 mode?
			yuv444torgb24( &convBlock ); // convert color y444 -> RGB24

		// emms resets FPU-stack state (necessary due to use of MMX 
		__asm emms; // emms *MUST* be called prior to calling avi-function
		// secure emms.. reintegrated px3 ver 0.1.0

		// store current frame to AVI writer
		store_avi_one( outname, fb_yuv, &CropRect, g_cbMPEGPicture ); 

	} // end if ( tgaflag == T_GDI )
	
  } // endif - if ( y_ptr != NULL ) 

  y_ptr = NULL;
}
/*
static void putbyte(c)
int c;
{
  *optr++ = c;

  if (optr == obfr+OBFRSIZE)
  {
    _write(outfile,obfr,OBFRSIZE);
    optr = obfr;
  }
}

static void putword(w)
int w;
{
  putbyte(w); putbyte(w>>8);
}

*/