/* storeavi.c : Windows AVIfile routines
 *
 * MPEG2AVI
 * --------
 *
 * avi_fileopen() - (private) init AVI routines, open AVI file
 * avi_store_picture() - (private) write one frame to AVI-file
 * store_avi_one() - (public) interface to store.c
      (called by store_avi_gdi() 
 * avi_fileclose() - (private) close AVI streams, close AVIfile

   This file was added to the original MSSG (MPEG Software Simulation
   Group) MPEG2V12 codec.  It enables the MPEG2V12 decoder to
   write an AVI file, using the Win32 AVI/VFW interface.
   The target machine must have the "Video Compression" system component
   installed on his Windows operating system.
   This file is *NOT* part of the original MSSG MPEG2V12 distribution!

 *
 *  Revision history
 *  v0.11 first release
 *  v0.12 added encoding-time statistic 
 *  v0.13 added abort-key ('Q') to close AVI-file and exit
 *  v0.14 added rectangular-clipping (centered) and 1/2 resolution output
          Data is inverse telecined prior to entering storeavi.c
		  Functions *depend* on proper-coordinates in
		  Resized_Picture_Height, Resized_Picture_Width, g_AVIcropX,
		  g_AVIcropY, CropRect
	v0.15 abort-key changed to #define KEY_QUIT (in global.h)
	      fixed elapsed-time report for hours:minutes
		  AVI-writer reports "MPEG FRAME # :", not framesWritten.
		  (framesWritten caused confusion with IVTC)
	v0.15B4 swtiched from writing to pAVIstream_compressed, to using
	      ICSeqCompressFrame() functions and writing to the uncompressed
		  AVIstream (pAVIstream)
		  A little more (programming) work on my part, but this new
		  code can directly compress YUV-bitmaps.
		  ICSeqCompressFrame() supports YUV-bitmaps, whereas
		  AVIStreamSetFormat() did not.  So far, only a handful of
		  codecs support YV12.
		  +avi_store_picyv12() ... writes YV12 bitmap to pAVIstream
		   (the regular avi_store_picture() writes to pAVIstream_compressed)
	v0.15B5 recoded the YV12-writer with Avery Lee's VirtualDub code
	      added YUY2 mode.  The YV12/YUY2 framebuffer formatting is handled
		  inside store.c (store_avi_gdi()), the functions here assume
		  the framebuffer has been properly prepared.
		  Finally corrected the code for setting g_iccompressframes's member 
		  variables.  The MPEG-4 KeyFrame-rate issue has been squashed!
		  Unfortunately, the compression bitrate control is now inaccurate (DOH)
		  put an if() statement around the ICCompressframeStart() code in
		    avi_fileopen()... fixes problem when RGB24 "Full Frames" is selected
	v0.15B51 finally the YUV-AVI mode's compressor setup is functioning
	      Turns out, to *properly* initialize the codec, this sequence must be
		  executed :
		    1) ICCompressorChoose();  // dialog-box, user chooses a codec
			2) ICGetState( hic, lpUsersSettings, cbState ); // store user's settings
			3) ICSetState( hic, NULL, cbState );    // reset the codec
			4) ICSetState( hic, lpUsersSettings, cbState ); //restore user's settings 

            The Microsoft MPEG-4 codecs require step #3 (reset codec)
		  also opened avi_fileopen() to the public.  It is no longer called by
		  avi_store_one(), it is now directly called by store_avi_gdi()

 */

#include<stdio.h>
#include<conio.h>
#include<string.h>
#include<time.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include "config.h"
#include "global.h"
#include<windows.h>
#include<vfw.h>

/* private functions */
// avi_fileopen() is now public
//int avi_fileopen( const char aviFilename[], LPRECT CropRect, DWORD dwPixFormat );
void avi_store_rgb( unsigned char *source_frame );
void avi_store_yuv( unsigned char *source_frame );
int get_fps( LPDWORD lpfpsRate, LPDWORD lpfpsScale );

DWORD fps_table[][2] = { 
	{24000, 1001},	/* [0] = 23.976fps NTSC film */
	{30000, 1001},  /* [1] = 29.97 fps NTSC video */
	{25   , 1   },  /* [2] = 25.000fps PAL film */
	{24   , 1   },  /* [3] = 24.000fps */
	{30   , 1   },  /* [4] = 30.000fps */
	{ -1  , -1  }  /* end of table */
};

typedef union tfccDword2Char
{
	DWORD dword;	/* the FCC code */
	BYTE ch[4]; /* character decomposition of dword */
} fccDword2Char;

time_t start_time; /* clock snapshot */

#define _AVIRATE_TABLELENGTH 5
#define MAX_FRAMECOUNT 0x7FFFFFFFL  /* max 31-bit positive value */


/*
typedef struct { 
    DWORD fccType; 
    DWORD fccHandler; 
    DWORD dwFlags; 
    DWORD dwCaps; 
    WORD  wPriority; 
    WORD  wLanguage; 
    DWORD dwScale; 
    DWORD dwRate; 
    DWORD dwStart; 
    DWORD dwLength; 
    DWORD dwInitialFrames; 
    DWORD dwSuggestedBufferSize; 
    DWORD dwQuality; 
    DWORD dwSampleSize; 
    RECT  rcFrame; 
    DWORD dwEditCount; 
    DWORD dwFormatChangeCount; 
    char  szName[64]; 
} AVISTREAMINFO; 
*/ 

/* private variables */

static PAVIFILE         g_pAVIfile; 
static PAVISTREAM       g_pAVIstream;
static PAVISTREAM		 g_pAVIstream_compressed;
static BITMAPINFOHEADER g_bmih; /* RGB bitmapinfoheader */
static BITMAPINFO		 g_bmi;
static LONG             g_lStreamSize; 
static LPVOID           g_lpOld; 
static LPVOID           g_lpNew; 
static LPVOID			g_lpVideoData;
static COMPVARS		g_compvars;
static AVISTREAMINFO    g_AVIstreaminfo; 
static AVICOMPRESSOPTIONS g_avicompressoptions;
static ICCOMPRESSFRAMES g_iccompressframes;  // for IC
static BITMAPINFO g_outbmi; // for IC

static LONG g_lFrameCount; // running counter of all frames written to 
							// AVI file, includes both key and non-key frames
static LONG g_lByteCount; // running counter of bytes written to AVI file
static LONG g_lIVTCCount; // running counter of MPEG frames deleted by IVTC
static LONG g_lKFCount; // running counter of key-frames written to AVI file

static unsigned char *g_lpCompBuf = NULL; // Compressor buffer, current frame
static unsigned char *g_lpCompBufP = NULL; // Compressor buffer, previous frame





// center a smaller DEST rectangle in the middle of a larger SRC rectangle
void
SetRectCenter( 
  LPRECT lprc, // pointer to structure with rectangle to set
  int src_x,   // X-dimension of source image
  int src_y,    // Y-dimension of source image
  int dest_x,  // X-dimension of destination image
  int dest_y)  // Y-dimension of destination image
{
	int mid_x =  (src_x - dest_x)>>1;
	int mid_y = (src_y - dest_y)>>1;

	SetRect( lprc, mid_x, mid_y, src_x - mid_x, src_y - mid_y );
}


/*
 * get_fps()  [ this code was originally part of avi_fileopen() ]
 * parse fps from user (or command-line option)
 */
int
get_fps( LPDWORD lpfpsRate, LPDWORD lpfpsScale )
{
	DWORD fpsRate = 0, fpsScale = 0;
	int i, ch;

	if ( g_AVIrate == _ASK_USER )
	{	/* g_AVIrate wasn't set by command-line so ask NOW */
		i = 0;
		do {
			printf("\n\t(%d) = %.3f fps   <%ld/%ld>", i, 
				(double)fps_table[ i ][ 0 ] / fps_table[ i ][ 1 ],
				fps_table[ i ][ 0 ] , fps_table[ i ][ 1 ] );
			i++;
		} while ( i < _AVIRATE_TABLELENGTH );
		printf("\n\t(%d) = Custom frame-rate ", _AVIRATE_TABLELENGTH );
		if ( CHECKOPT( UserOptIVTC ) ) /* IVTC in use? */
			printf("\n\t(IVTC active, recommendation is 23.976 or 24 fps)");
		printf("\n\tSelect AVI frame-rate : ");
		scanf( " %d", &g_AVIrate );
	}

	if ( g_AVIrate < 0 || g_AVIrate > _AVIRATE_TABLELENGTH )
	{
		fprintf( stderr, "\nERROR : %d is an invalid frame-rate selection!",
			g_AVIrate );
		return ERROR;
	}

	if ( g_AVIrate < _AVIRATE_TABLELENGTH )
	{
		printf( " \nAVI frame rate = %.3f (fps), ", (double) 
			fps_table[ g_AVIrate ][ 0 ] / fps_table[ g_AVIrate ][ 1 ] );
		fpsScale = fps_table[ g_AVIrate ][ 1 ];
		fpsRate = fps_table[ g_AVIrate ][ 0 ];
	}
	else 
	{	/* 0.15B2 custom AVI-frame rate selection */
	  printf("\nCustom selection : fps = ( dwRate / dwScale )" );
	  do {
		printf("\nChoose dwRate (decimal) : " );
		scanf( " %ld", &fpsRate );
		printf("Choose dwScale (decimal) : " );
		scanf( " %ld", &fpsScale );
		printf("frame rate = %.3f (fps)  ... Acceptable?", (double) fpsRate /
			fpsScale );
		ch = _getch();
		putchar ( toupper( ch ) );
	  } while ( toupper( ch ) != 'Y' );
	  printf("\n" );
	}

	*lpfpsRate = fpsRate;
	*lpfpsScale = fpsScale;
	return 1; // OK, != ERROR
} // endfunction get_fps()


/*
 * avi_fileopen()
 * opens a new file for writing
 * Queries user for Compressor (VCM) selection
 * Queries user to choose frame-rate (dwRate, dwScale)
 */
int
avi_fileopen( const char aviFilename[], LPRECT CropRect, DWORD dwPixFormat )
{
    HRESULT          hr; 
	FILE	*fp;
	BOOL brc = FALSE; /* boolean return code */
	DWORD dwrc = 0; /* DWORD return code */
	char ch, text[ 256];
	unsigned char *lpState = NULL; /* temporary storage area for codec state */
	int i = 0;
	fccDword2Char fccD2C; /* custom struct for converting DWORD -> char[4] */
	DWORD fpsScale, fpsRate; /* user-input vars for AVI frame-rate selection */
	LPAVICOMPRESSOPTIONS m_lpavicompressoptions = &g_avicompressoptions;
	//COMPVARS m_compvars2;
	BITMAPINFOHEADER m_bmih_yuv, m_bmih_rgb24;
	ICINFO m_icinfo;
	BOOL fUseYUV = (dwPixFormat != BI_RGB); /* use m_bmih_yuv to setup AVI? */

	//-------- 0) initialize AVI counters, and AVI library routines
	g_lFrameCount = g_lByteCount = g_lIVTCCount = g_lKFCount = 0;
	AVIFileInit();

	// Initialize the bitmap structure. 
	m_bmih_rgb24.biSize = sizeof(BITMAPINFOHEADER); 
	// 0.14 The yuv2rgb24() convertor crops the output to dimensions CropRect,
	// thus the final frame's size is determined by CropRect 
	m_bmih_rgb24.biWidth = CropRect->right - CropRect->left ;
	m_bmih_rgb24.biHeight = CropRect->bottom - CropRect->top ;
	// some codecs *require* valid Height/Width params to avoid ICERR_BADFORMAT
	m_bmih_rgb24.biPlanes = 1;			// must be 1
	m_bmih_rgb24.biCompression = BI_RGB;	// standard RGB bitmap
	m_bmih_rgb24.biBitCount = 24;		// 8 bits-per-pixel R,G,B format 
	//g_bmih.biSizeImage = 0;			// okay to be 0 for BI_RGB
	m_bmih_rgb24.biSizeImage = m_bmih_rgb24.biWidth*
		m_bmih_rgb24.biHeight * 3;
	m_bmih_rgb24.biXPelsPerMeter = m_bmih_rgb24.biYPelsPerMeter = 0;
	m_bmih_rgb24.biClrUsed = m_bmih_rgb24.biClrImportant = 0;

	// set YUV version of BITMAPINFOHEADER
	// note, "IYUV" and "I420" are the same as YV12, but these other two
	// FOURCCs cause "full frames" to be listed, which doesn't work.
	// so we stick with setting biCompression = 'YV12'
	m_bmih_yuv = m_bmih_rgb24;
	//m_bmih_yuv.biCompression = mmioFOURCC( 'Y', 'V', '1', '2' );
	m_bmih_yuv.biCompression = dwPixFormat;
	m_bmih_yuv.biSizeImage = m_bmih_yuv.biWidth* m_bmih_yuv.biHeight;
	if ( fUseYUV )
	{
	  switch ( dwPixFormat )
	  {
		case mmioFOURCC( 'Y','V','1','2') : 
			m_bmih_yuv.biBitCount = 12;
			m_bmih_yuv.biSizeImage = ( m_bmih_yuv.biSizeImage * 3 ) >> 1;
			break;
		case mmioFOURCC( 'I','4','2','0') :
			m_bmih_yuv.biBitCount = 12;
			m_bmih_yuv.biSizeImage = ( m_bmih_yuv.biSizeImage * 3 ) >> 1;
			break;
		case mmioFOURCC( 'I','Y','U','V') :
			m_bmih_yuv.biBitCount = 12;
			m_bmih_yuv.biSizeImage = ( m_bmih_yuv.biSizeImage * 3 ) >> 1;
			break;
		case mmioFOURCC( 'Y','U','Y','2') :
			m_bmih_yuv.biBitCount = 16;
			m_bmih_yuv.biSizeImage = m_bmih_yuv.biSizeImage << 1;
			break;
		case BICOMP_CREATIVEYUV:
			m_bmih_yuv.biBitCount = 16;
			m_bmih_yuv.biSizeImage = m_bmih_yuv.biSizeImage << 1;
			break;
		default :
			fccD2C.dword = dwPixFormat;
			fprintf( stderr, "\navi_fileopen() : unrecognized pixFormat FOURCC '%c%c%c%c' !",
				fccD2C.ch[0], fccD2C.ch[1], fccD2C.ch[2], fccD2C.ch[3] );
			exit( ERROR );
	  } /* endswitch */
	} /* endif ( fUseYUV ) */

	// set global g_bmih variable
	g_bmih = ( fUseYUV ? m_bmih_yuv : m_bmih_rgb24 );
	
	printf("\nOutput AVI-resolution : <%d X %d> : pixFormat = ", 
		m_bmih_rgb24.biWidth,m_bmih_rgb24.biHeight );
	if ( fUseYUV )
	{
		fccD2C.dword = dwPixFormat;
		printf( "'%c%c%c%c'", fccD2C.ch[0], fccD2C.ch[1], fccD2C.ch[2], 
			fccD2C.ch[3] );
	}
	else
	{
		printf( "RGB24");
	} /*endif ( fUseYUV ) */

	if ( (m_bmih_rgb24.biWidth & 0x4) || (m_bmih_rgb24.biHeight & 0x4) )
		printf("\n***WARNING*** chosen viewport dimension(s) are not DIV 16!");


 	//-------- 1)  Ask user to choose a Video Compression Codec
	g_compvars.cbSize = sizeof( g_compvars ); //  must initialize cbSize!
	if ( fUseYUV )
	{
		sprintf( text, "ICCompressorChoose() : select codec %c%c%c%c",
			fccD2C.ch[0], fccD2C.ch[1], fccD2C.ch[2], fccD2C.ch[3] );
		brc = ICCompressorChoose( NULL, ICMF_CHOOSE_DATARATE | 
			ICMF_CHOOSE_KEYFRAME, &m_bmih_yuv, NULL, &g_compvars, text );
	}
	else
		brc = ICCompressorChoose( NULL, ICMF_CHOOSE_DATARATE | 
			ICMF_CHOOSE_KEYFRAME , &m_bmih_rgb24, NULL, &g_compvars,
			"ICCompressorChoose() : select codec RGB24" );

	g_compvars.lpbiIn = &g_bmi; // this is sometimes not set 

	if ( brc == 0 )	
	{
		fprintf( stderr, "\navi_fileopen() : Video codec selection failed!" );
		exit(ERROR);
	}
	else
	{	/* display the chosen codec's FOURCC */
		// get info from codec, we'll use this info later for g_iccompressframes
		dwrc = ICGetInfo( g_compvars.hic, &m_icinfo, sizeof(m_icinfo) );
		if ( !dwrc )
		{
			fprintf( stderr, "\navi_fileopen() : ICGetInfo() failed.");
			exit( ERROR );
		}

		fccD2C.dword = g_compvars.fccHandler;
		printf( "\n\tAVI Video Compressor fccHandler='%c%c%c%c' ", 
			fccD2C.ch[0], fccD2C.ch[1], fccD2C.ch[2], fccD2C.ch[3] );

	} /* endif ( brc == 0 ) */

	// allocate ram for state
	lpState = (unsigned char *)malloc( g_compvars.cbState );
	if ( lpState == NULL )
	{
		fprintf( stderr, "\navi_fileopen() : malloc() failed.");
		exit( ERROR );
	}


	// reset the codec, very important! 
	ICGetState( g_compvars.hic, lpState, g_compvars.cbState );
	ICSetState( g_compvars.hic, NULL, g_compvars.cbState );
	ICSetState( g_compvars.hic, lpState, g_compvars.cbState );
	
	//-------- 2) check if file already exists, if so prompt user to overwrite
	fp = fopen( aviFilename, "rb");
	if ( fp != NULL )
	{
		fclose( fp );
		printf("\nOverwrite %s ? ", aviFilename );
		ch = getchar();
		if ( ch != 'y' && ch != 'Y' )
			exit( ERROR );

		fprintf( stderr, "Overwriting %s!", aviFilename );
		
		fp = fopen( aviFilename, "wb" );	// set filelength=0
		if ( fp == NULL )
		{
			fprintf( stderr, "\navi_fileopen() : Could not overwrite %s!", aviFilename );
			exit( ERROR );
		}

		fclose( fp );
	}


    //-------- 3) Create our AVI-file using AVIFileOpen(). 
    hr = AVIFileOpen(&g_pAVIfile, aviFilename, OF_WRITE | OF_CREATE, NULL); 
    if (hr != 0)
	{
		fprintf( stderr, "\navi_fileopen() : AVIFileOpen() couldn't create %s!\n", aviFilename );
		switch ( hr ) {
		case AVIERR_BADFORMAT :
			fprintf( stderr, "The file couldn't be read, indicating a corrupt file or an unrecognized format.");
			break;
		case AVIERR_MEMORY:
			fprintf( stderr, "The file could not be opened because of insufficient memory.");
			break;
		case AVIERR_FILEREAD :
			fprintf( stderr, "A disk error occurred while reading the file.");
			break;
		case AVIERR_FILEOPEN :
			fprintf( stderr, "A disk error occurred while opening the file. ");
			break;
		case REGDB_E_CLASSNOTREG:
			fprintf( stderr, "According to the registry, the type of file specified in AVIFileOpen does not have a handler to process it.");
			break;
		default:
			fprintf( stderr, "Unknown return code %d", hr );
		}
		ICCompressorFree( &g_compvars ); 
        exit(ERROR);
	}


	//-------- 4) Process AVI frame-rate
	i = get_fps( &fpsRate, &fpsScale );
	if ( i == ERROR )
	{
		ICCompressorFree( &g_compvars ); 
        AVIFileRelease( g_pAVIfile ); 
		exit( ERROR );
	}		


	//-------- 5) load AVIstreaminfo structure, lot of busy work!
	g_AVIstreaminfo.fccType = streamtypeVIDEO; 
    g_AVIstreaminfo.fccHandler = g_compvars.fccHandler;
	g_AVIstreaminfo.dwFlags = 0; 
    g_AVIstreaminfo.dwCaps =0;// unused, must be 0
    g_AVIstreaminfo.wPriority =0 ; // ?
    g_AVIstreaminfo.wLanguage =0;  // ?
	g_AVIstreaminfo.dwScale = fpsScale; // see step 4)
	g_AVIstreaminfo.dwRate = fpsRate;   // see step 4)
	g_AVIstreaminfo.dwStart = 0; 
    g_AVIstreaminfo.dwLength = 0; 
    g_AVIstreaminfo.dwInitialFrames = 0; // no audio/video skewing
    g_AVIstreaminfo.dwSuggestedBufferSize = g_bmih.biSizeImage; 
    g_AVIstreaminfo.dwQuality = -1; // default
    g_AVIstreaminfo.dwSampleSize = 0; // variable
	SetRect(&g_AVIstreaminfo.rcFrame, 0, 0,(int) m_bmih_rgb24.biWidth, 
		(int) m_bmih_rgb24.biHeight );
	g_AVIstreaminfo.dwEditCount = 0; 
    g_AVIstreaminfo.dwFormatChangeCount = 0; 
		/* create stream header text */
    strcpy( g_AVIstreaminfo.szName, Version ); /* Version = "mpeg2avi ..." */
	sprintf( text, " %s-IDCT, ", Reference_IDCT_Flag ? "64FPU" : "32INT" );
	strcat( g_AVIstreaminfo.szName, text );
	sprintf( text, "%s filter", g_fastfilter ? "Fast" : "Hi-Q" );
	strcat( g_AVIstreaminfo.szName, text );
	if ( CHECKOPT( UserOptIVTC ) ) /* is IVTC in use? */
	{
		sprintf( text, ", -s%d IVTC(%d)", GETOPT( UserOptIVTC ), 
			GETOPTAUX( UserOptIVTC )  ); /* extended IVTC config info */
		strcat( g_AVIstreaminfo.szName, text );
	}

    //-------- 6) Create a stream using AVIFileCreateStream. 
    hr = AVIFileCreateStream(g_pAVIfile, &g_pAVIstream, &g_AVIstreaminfo); 
    if (hr != ICERR_OK) {     //Stream created OK? If not, close file. 
		ICCompressorFree( &g_compvars ); 
        AVIFileRelease( g_pAVIfile ); 
		fprintf( stderr, 
			"\navi_fileopen() : AVIFileCreateStream() : Couldn't create stream!");
        exit(ERROR); 
    } 

	//-------- 7) AVI Compression-stream setup
	// prior to 0.15B4, we used AVISaveOptions() to select the codec.
	// but now, ICCompressorChoose() [step 1] replaces AVISaveOptions(), so 
	// g_avicompressoptions must be manually set-up.
	// 
	// The translation process g_compvars -> g_avicompressoptions is 100%
	// accurate, but that's ok.  Some testing showed the translation to be
	// good enough (i.e. no AVI errors.)

	g_avicompressoptions.cbParms = g_compvars.cbState;
	g_avicompressoptions.cbFormat = sizeof( g_bmih );  // critical!
	// g_compvars.lDataRate is in units of KBPS : KBPS = 1024*BPS 
	g_avicompressoptions.dwBytesPerSecond = (g_compvars.lDataRate * 1024);
	g_avicompressoptions.dwFlags=AVICOMPRESSF_VALID | AVICOMPRESSF_KEYFRAMES
		| AVICOMPRESSF_DATARATE; // AVICOMPRESSF_VALID is critical!
	g_avicompressoptions.dwInterleaveEvery = 0; // no interleaving...
	g_avicompressoptions.dwKeyFrameEvery = g_compvars.lKey;
	g_avicompressoptions.dwQuality = g_compvars.lQ;
	g_avicompressoptions.fccHandler = g_compvars.fccHandler; // critical!
	g_avicompressoptions.fccType = streamtypeVIDEO; // critical!
	g_avicompressoptions.lpFormat = &g_bmih; // critical!
	//g_avicompressoptions.lpParms = g_compvars.lpState;
	
	// make a 2nd copy of the g_compvars.lpState data
	g_avicompressoptions.lpParms = (LPVOID) malloc ( g_compvars.cbState);
	if ( g_avicompressoptions.lpParms == NULL )
	{
		ICCompressorFree( &g_compvars ); 
		AVIStreamRelease(g_pAVIstream); 
        AVIFileRelease( g_pAVIfile ); 
		fprintf( stderr, "\navi_fileopen() : malloc() failed.");
		exit( ERROR );
	}
	else
		memcpy( g_avicompressoptions.lpParms, g_compvars.lpState, 
			g_compvars.cbState);

	// v0.15B4 no longer used..., now using ICCompressorChoose() 
	// Display Compression-codec dialog-box.  Get AVICOMPRESSOPTIONS
	/*
	if (!AVISaveOptions(NULL, ICMF_CHOOSE_KEYFRAME | ICMF_CHOOSE_DATARATE
		//| ICMF_CHOOSE_PREVIEW, 1, &g_pAVIstream, (LPAVICOMPRESSOPTIONS FAR *) 
		, 1, &g_pAVIstream, (LPAVICOMPRESSOPTIONS FAR *) 
		&m_lpavicompressoptions)) 
	{
		fprintf( stderr, "\nAVISaveOptions() : failed!");
		exit(ERROR);
	}
	else
	{
		fccD2C.dword = g_avicompressoptions.fccHandler;
		printf( "\nAVICOMPRESSOPTIONS.fccHandler='%c%c%c%c'", fccD2C.ch[0], 
			fccD2C.ch[1], fccD2C.ch[2], fccD2C.ch[3] );
		printf(" lKey = %ld", g_avicompressoptions.dwKeyFrameEvery );
	}*/


	//-------- 9) make compressed stream
	// To write a compressed frame using the user selected codec, set
	// format with AVIStreamSetFormat(), then write with
	// AVIStreamWrite( g_pAVIstream_compressed, ... )
	hr = AVIMakeCompressedStream(&g_pAVIstream_compressed, g_pAVIstream,
		&g_avicompressoptions, NULL); 
	if (hr != AVIERR_OK)
	{ 
		fprintf( stderr, "\nAVIMakeCompressedStream() : failed! " );
		switch( hr ) {
			case AVIERR_NOCOMPRESSOR :
				fprintf( stderr, "A suitable compressor cannot be found." );
				break;
			case AVIERR_MEMORY :
				fprintf( stderr, "There is not enough memory to complete the operation.");
				break;
			case AVIERR_UNSUPPORTED :
				fprintf( stderr, "Compression is not supported for this type of data." );
				fprintf( stderr, "\n\tThis error might be returned if you try to compress" );
				fprintf( stderr, "\n\tdata that is not audio or video. ");
				break;
			default:
				fprintf( stderr, "Unknown return code %d", hr );
		}
		ICCompressorFree( &g_compvars ); 
		AVIStreamRelease(g_pAVIstream); 
        AVIFileRelease(g_pAVIfile); 
		exit(ERROR);
	}


	//-------- 10) Set the format of the AVISTREAM object
	// Set format of compressed stream using AVIStreamSetFormat.
    // Note, AVIStreamSetFormat() can *only* be called ONCE 
	// for an pAVIfile.  The 2nd attempt will always fail!
	// Thus, if two pAVIstreams point to the same pAVIfile,
	// SetFormat() can only be called once between these
	// two pAVIstreams.
	
	// v0.11 corrected bug, where the wrong size-info was
	// passed to AVIStreamSetFormat().  The bug caused AVI files written
	// in UNCOMPRESSED mode to report the incorrect color-mask table.
	//hr = AVIStreamSetFormat(g_pAVIstream_compressed,0,&m_bmih_rgb24,
	hr = AVIStreamSetFormat(g_pAVIstream_compressed,0,&m_bmih_rgb24,
		sizeof(m_bmih_rgb24) );
    if (hr != ICERR_OK ) { 
		fprintf( stderr, "\nAVIStreamSetFormat() : Couldn't set format!");
		ICCompressorFree( &g_compvars );
		AVIStreamRelease(g_pAVIstream_compressed); 
        AVIStreamRelease(g_pAVIstream); 
        AVIFileRelease(g_pAVIfile); 
        exit( ERROR ); 
    }

	// correct Microsoft MPEG-4's bogus keyframe-rate report
	// setting lKeyRate = 0 disables MPEG2AVI's keyrate regulator

  if ( fUseYUV )
  {
	// setup BITMAPINFO structure for ICSeqCompressFrameStart()
	g_bmi.bmiHeader = g_bmih;
	//g_bmi.bmiHeader.biBitCount =32;
	g_bmi.bmiColors[0].rgbBlue=g_bmi.bmiColors[0].rgbRed=
		g_bmi.bmiColors[0].rgbGreen=g_bmi.bmiColors[0].rgbReserved = 0;

	if ( ICCompressGetFormat( g_compvars.hic, &g_bmi, &g_outbmi ) != ICERR_OK )
	{
		fprintf(stderr, "\nICCompressGetFormat() failed!\n");
	}

	// reset BITMAPINFO structure for AVIStreamSetFormat & ICSeqCompressFrame
	g_bmi.bmiHeader = g_bmih;
	g_bmi.bmiColors[0].rgbBlue=g_bmi.bmiColors[0].rgbRed=
		g_bmi.bmiColors[0].rgbGreen=g_bmi.bmiColors[0].rgbReserved = 0;

	g_iccompressframes.dwFlags =0;
	g_iccompressframes.lpbiOutput = &g_outbmi.bmiHeader;
	g_iccompressframes.lpbiInput = &g_bmih;
	g_iccompressframes.lOutput = 0; // reserved
	g_iccompressframes.lInput = 0; // reserved
	g_iccompressframes.lStartFrame = 0;	// start from frame#0
	g_iccompressframes.lFrameCount = MAX_FRAMECOUNT;  // # frames to compress
	if ( (m_icinfo.dwFlags & VIDCF_QUALITY) )
		g_iccompressframes.lQuality = g_compvars.lQ;
	else
		g_iccompressframes.lQuality = 0; /* disable Quality parameter */
	g_iccompressframes.lDataRate = (g_compvars.lDataRate*1024);
	if ( g_compvars.lKey == 9999 )
		g_iccompressframes.lKeyRate = 0; // disable key-frame regulator
	else
		g_iccompressframes.lKeyRate = g_compvars.lKey;
	//g_iccompressframes.dwRate = 1000000;  // # microseconds in 1 second
	//g_iccompressframes.dwScale= 1000000.0 / ( (double) fpsRate / fpsScale );
	g_iccompressframes.dwRate = fpsRate;   // # microseconds in 1 second
	g_iccompressframes.dwScale= fpsScale;
	g_iccompressframes.dwOverheadPerFrame = 0; // reserved
	g_iccompressframes.dwReserved2 = 0; // reserved
	g_iccompressframes.GetData = NULL; // not in use
	g_iccompressframes.PutData = NULL; // not in use

	//ICGetState( g_compvars.hic, text, g_compvars.cbState );
	// must send ICSendMessage() before calling ICSeqComrpessFrameStart()
	ICSendMessage( g_compvars.hic, ICM_COMPRESS_FRAMES_INFO, (WPARAM)
		&g_iccompressframes, (DWORD) sizeof( g_iccompressframes ) );
	//ICSetState( g_compvars.hic, text, g_compvars.cbState );

	if ( !ICSeqCompressFrameStart( &g_compvars, &g_bmi )  )
	{
		fprintf( stderr, "\nICSeqCompressFrameStart() : failed!" );
		ICCompressorFree( &g_compvars ); 
		AVIStreamRelease(g_pAVIstream_compressed); 
		AVIStreamRelease(g_pAVIstream); 
		AVIFileRelease(g_pAVIfile); 
		exit(ERROR);
	}
  } /* endif ( fUseYUV ) */

	time( &start_time );	/* Starting counting the clock NOW */
	return TRUE;	/* function is successful */
} // endfunction avi_fileopen()


void
avi_store_rgb24( unsigned char *source_frame )
{
	BOOL frc; // boolean return code;
	BOOL fIsKeyFrame= 0; // set by ICSeqCompressFrame, key frame?
	DWORD dwFlagsKF = 0; // flag for keyframe

	LONG lSamplesWritten;	// #frames written, return parameter
	LONG lBytesWritten;
	//LONG lCompVideoDataSize;
	static BYTE lpstateBuff[ 256 ];
	
	LPVOID lpCompVideoData = NULL;	// compressed video data

    // Save the compressed data using AVIStreamWrite(). 
    // The AVIStreamWrite function writes data to a stream.
	frc = AVIStreamWrite( g_pAVIstream_compressed, g_lFrameCount, 1, 
		source_frame, g_bmih.biSizeImage, 0, 
		&lSamplesWritten, &lBytesWritten) ;
	if ( frc )
	{
		fprintf( stderr, "\navi_store_rgb24() : AVIStreamWrite() failed!");
	}
	else
	{
		// update global counters 
		g_lFrameCount += lSamplesWritten;	// count of all frames
		g_lByteCount += lBytesWritten;		// count of bytes
		g_lKFCount += (fIsKeyFrame?1:0);	// count of key-frames
		printf("AVIStreamWrite() (%ld) bytesWritten=%ld %s",
			g_lByteCount, lBytesWritten, (fIsKeyFrame ? "KEYF  ":"      ") );
	}
}

void
avi_store_yuv( unsigned char *source_frame )
{
	BOOL frc; // boolean return code;
	BOOL fIsKeyFrame= 0; // set by ICSeqCompressFrame, key frame?
	DWORD dwFlagsKF = 0; // flag for keyframe
//	static DWORD dwLastKF = 0; // frame# of last keyframe 

	LONG lSamplesWritten;	// #frames written, return parameter
	LONG lBytesWritten, lCompVideoDataSize;
	static BYTE lpstateBuff[ 256 ];
	
	LPVOID lpCompVideoData = NULL;	// compressed video data

	//if ( g_lFrameCount == 0 )
	//	ICGetState( g_compvars.hic, lpstateBuff, g_compvars.cbState );

	//ICSetState( g_compvars.hic, lpstateBuff, g_compvars.cbState );
	lpCompVideoData = ICSeqCompressFrame( &g_compvars, 0, 
		(LPVOID) source_frame, &fIsKeyFrame, &lCompVideoDataSize );
	if ( lpCompVideoData == NULL )
	{
		fprintf( stderr, "\navi_store_yuv() : ICSeqCompressFrame() failed!" );
		return;
	}
	else
	{
		//printf(" ICSeq() : CompSize=%ld, fr=%ld, %s", lCompVideoDataSize,
		//	g_compvars.lFrame, (fIsKeyFrame ? "KEYF " : "     "));
		// convert fIsKeyFrame to proper dwFlags for AVIStreamWrite() 
		dwFlagsKF = (fIsKeyFrame ? AVIIF_KEYFRAME : 0);
		//ICGetState( g_compvars.hic, lpstateBuff, g_compvars.cbState );
		//g_lByteCount += lCompVideoDataSize;		// count of bytes
		//g_lKFCount += (fIsKeyFrame?1:0);	// count of key-frames
		//printf( "total=%ld (%ld)  ", g_lByteCount, g_lKFCount );
		//++g_lFrameCount;
	}

	/*if ( g_lFrameCount == 0 )
  {
	frc = AVIStreamWrite( g_pAVIstream_compressed, g_lFrameCount, 1, 
		source_frame, g_bmih.biSizeImage, 0, 
		&lSamplesWritten, &lBytesWritten) ;
	++g_lFrameCount;
	if ( frc )
		fprintf( stderr, "\nAVIStreamWrite() failed!\n");
	else
		ICSetState( g_compvars.hic, g_compvars.lpState, g_compvars.cbState );
  }*/
 // else


    // Save the compressed data using AVIStreamWrite(). 
    // The AVIStreamWrite function writes data to a stream.
	
	//frc = AVIStreamWrite( g_pAVIstream_compressed, g_lFrameCount, 1, 
	//	source_frame, g_bmih.biSizeImage, 0, 
	if ( lpCompVideoData )
		frc = AVIStreamWrite( g_pAVIstream, g_lFrameCount, 1, 
			lpCompVideoData, lCompVideoDataSize, dwFlagsKF, 
			&lSamplesWritten, &lBytesWritten) ;
	if ( frc )
	{
		fprintf( stderr, "\navi_store_yuv() : AVIStreamWrite() failed!");
	}
	else
	{
		if ( lCompVideoDataSize != lBytesWritten )
			fprintf( stderr, 
				"\n#bytes written by AVIStreamWrite()!=ICSeqCompressFrame()");
		// update global counters 
		g_lFrameCount += lSamplesWritten;	// count of all frames
		g_lByteCount += lBytesWritten;		// count of bytes
		g_lKFCount += (fIsKeyFrame?1:0);	// count of key-frames
		printf("AVIStreamWrite() (%ld) bytesWritten=%ld %s",
			g_lByteCount, lBytesWritten, (fIsKeyFrame ? "KEYF  ":"      ") );
	//	if ( fIsKeyFrame )
	//	{
	//		printf (" %ld\n", g_lFrameCount - dwLastKF );
	//		dwLastKF = g_lFrameCount;
	//	}
	}
}



void
avi_fileclose( )
{
	time_t finish_time, el_time; /* elapsed time */
	LPAVICOMPRESSOPTIONS m_lpavicompressoptions = &g_avicompressoptions;
	BOOL fUseYUV = ( Output_Type != T_AVI );

	/* clean up code, release resources for requested system services */
	ICSeqCompressFrameEnd( &g_compvars ) ;
	ICCompressorFree( &g_compvars ); 

	AVISaveOptionsFree(1,(LPAVICOMPRESSOPTIONS FAR *)&m_lpavicompressoptions);
	AVIStreamRelease(g_pAVIstream_compressed); 
    AVIStreamRelease(g_pAVIstream); 
    AVIFileRelease(g_pAVIfile); 

	time( &finish_time );	/* Record the finish time */

	/* time code copied from John Schlichther's AVI2MPG1 */
	el_time = finish_time - start_time;
	puts("");	/* new line */
	if(el_time <= 60)
		printf( "Elapsed time = %i seconds\n", el_time);
	else if(el_time < 3600)
		printf( "Elapsed time = %i:%02i minutes\n", (el_time/60)+1,
			el_time%60 );
	else
		printf( "Elapsed time = %i:%02i hours\n", el_time/3600, 
			(el_time%3600)/60);

	/* the key-frame data is only returned by ICSeqCompressFrame()...thus
	   RGB24-writing mode does not have access to key-frame info! */
	if ( fUseYUV ) 
		printf("\nAVI Statistics :  (Key)/Frames written=(%ld)/%ld, Bytes written=%ld",
			g_lKFCount, g_lFrameCount, g_lByteCount );
	else
		printf("\nAVI Statistics :  Frames written=%ld, Bytes written=%ld",
			g_lFrameCount, g_lByteCount );


	printf( "\nEstimated AVI video bitrate : %.1f KBPS", (double)
		(g_lByteCount>>10) * ( g_AVIstreaminfo.dwRate / 
		g_AVIstreaminfo.dwScale ) / g_lFrameCount );
	printf( "  ( @%.3f fps )", (double) g_AVIstreaminfo.dwRate / 
		g_AVIstreaminfo.dwScale );
	if ( CHECKOPT( UserOptIVTC ) )
		printf("\nInverse-telecine : start-state = %d ", 
		  GETOPT( UserOptIVTC ) );
}



/* this function is called by store.c, the rest of the functions in this
   file are private  */
void store_avi_one( const char aviFilename[], unsigned char *source_frame,
	LPRECT CropRect, long FrameNumber )
{
	static first_iteration = 1;
	static unsigned char *ivtc_frame; /* output frame from IVTC algorithm */
	int key;
	//BOOL fr;  /* boolean return code */
	BOOL fUseYUV = FALSE;
	DWORD dwPixFormat = 0; /* framebuffer pixel format, BI_RGB, YUV, etc. */

	switch ( Output_Type )
	{
	  case T_AVI_YV12 :
		fUseYUV = TRUE;
		dwPixFormat = mmioFOURCC( 'Y','V','1','2');
		break;
	  case T_AVI_YUY2 :
		fUseYUV = TRUE;
		dwPixFormat = mmioFOURCC( 'Y','U','Y','2');
		break;
	  default :
		dwPixFormat = BI_RGB;
		fUseYUV = FALSE;
		break;
	}  /* endswitch ( Output_Type ) */


	if ( first_iteration )
	{
		first_iteration = 0;	/* disable status flag */
		/* moved to store_avi_gdi() [ store.c]
		//fr = avi_fileopen( aviFilename, CropRect, dwPixFormat ); /* Init AVI file */
		//if ( !fr )
		//{
		//	fprintf( stderr, "\nCould not open %s for writing.", aviFilename );
		//	exit(ERROR);
		//}
		printf("\n<Q> to abort program execution (do not use CTRL-C)\n\n");
	}

	/* storeavi.c assumes that the source_frame[] is already in the correct
	   format (dwPixFormat.)  This routine does not check the correctness
	   of the bitmap's format. Any conversion must take place *PRIOR* to
	   calling this function, store_avi_one() */
	ivtc_frame = source_frame; /* already telecined in store.c */ 
	//printf("\rMPEG frame %ld : ", FrameNumber );

	if ( ivtc_frame != NULL )	/* have something to write? */
	{
		if ( fUseYUV )
			avi_store_yuv( ivtc_frame );  // write YUV frame
		else
		{
			/* convert YUV444 24bpp packed-pixel frame to RGB24 */
			yuv2rgb24( ivtc_frame, CropRect );
			avi_store_rgb24( ivtc_frame );	// write RGB24 frame
		}
	}
	else
	{	/* actually, IVTC is now performed inside the store.c */
		/* so store_avi_one() should never receive ivtc_frame == NULL */
		printf("NOT WRITTEN (IVTC)                                    ");
		++g_lIVTCCount;
	}

	/* v0.15B51, parse keyboard & previous QUIT command */
	if ( _kbhit() )	/* recent keypress?  If not, don't _getch() */
	{
		key = _getch();	/* yes, read keypress */
		if ( toupper( key ) == toupper( KEY_QUIT ) )
		{
			g_cmd_buttonStop = TRUE;	/* set command to stop */
		}
	}

	if ( g_cmd_buttonStop )
	{
		printf("\nDecoder stopped.");
		avi_fileclose();
		exit( ERROR );
	}

}



// unused
void
avi_store_yv12_old( unsigned char *source_frame )
{
	BOOL frc; // boolean return code;
	BOOL fIsKeyFrame= 0; // set by ICSeqCompressFrame, key frame?
	DWORD dwFlagsKF = 0; // flag for keyframe

	LONG lSamplesWritten;	// #frames written, return parameter
	LONG lBytesWritten, lCompVideoDataSize;
	static BYTE lpstateBuff[ 256 ];
	
	LPVOID lpCompVideoData = NULL;	// compressed video data

	// Compress the source_frame[] ... 
	// ICSeqCompressFrame() is called once for each frame in the input sequence.
	// Function returns NULL (fail), or pointer-to-data (success.)
	/*if ( !ICSetState( g_compvars.hic, lpstateBuff, 
		g_compvars.cbState ) )
	{
		fprintf( stderr, "\nICSetState() : failed!" );
	}*/

	g_iccompressframes.lStartFrame = g_compvars.lFrame;	// start from current frame
	//ICSendMessage( g_compvars.hic, ICM_COMPRESS_FRAMES_INFO, 
	//	(DWORD) sizeof( g_iccompressframes ), (DWORD) (LPVOID)
	//	&g_iccompressframes  );
	//ICSendMessage( g_compvars.hic, ICM_COMPRESS_FRAMES_INFO, 
	//	(DWORD) (LPVOID) &g_iccompressframes,
	//	(DWORD) sizeof( g_iccompressframes ) );

	lpCompVideoData = ICSeqCompressFrame( &g_compvars, 0, 
		(LPVOID) source_frame, &fIsKeyFrame, &lCompVideoDataSize );
	if ( lpCompVideoData == NULL )
	{
		fprintf( stderr, "\navi_store_picyv12() : ICSeqCompressFrame() failed!" );
		return;
	}
	else
	{
		//printf(" ICSeq() : CompSize=%ld, fr=%ld, %s", lCompVideoDataSize,
		//	g_compvars.lFrame, (fIsKeyFrame ? "KEYF " : "     "));
		// convert fIsKeyFrame to proper dwFlags for AVIStreamWrite() 
		dwFlagsKF = (fIsKeyFrame ? AVIIF_KEYFRAME : 0);
		//ICGetState( g_compvars.hic, lpstateBuff, g_compvars.cbState );
		//g_lByteCount += lCompVideoDataSize;		// count of bytes
		//g_lKFCount += (fIsKeyFrame?1:0);	// count of key-frames
		//printf( "total=%ld (%ld)  ", g_lByteCount, g_lKFCount );
		//++g_lFrameCount;
	}

	/*if ( g_lFrameCount == 0 )
  {
	frc = AVIStreamWrite( g_pAVIstream_compressed, g_lFrameCount, 1, 
		source_frame, g_bmih.biSizeImage, 0, 
		&lSamplesWritten, &lBytesWritten) ;
	++g_lFrameCount;
	if ( frc )
		fprintf( stderr, "\nAVIStreamWrite() failed!\n");
	else
		ICSetState( g_compvars.hic, g_compvars.lpState, g_compvars.cbState );
  }*/
 // else


    // Save the compressed data using AVIStreamWrite(). 
    // The AVIStreamWrite function writes data to a stream.
	
	//frc = AVIStreamWrite( g_pAVIstream_compressed, g_lFrameCount, 1, 
	//	source_frame, g_bmih.biSizeImage, 0, 
	if ( lpCompVideoData )
		frc = AVIStreamWrite( g_pAVIstream, g_lFrameCount, 1, 
			lpCompVideoData, lCompVideoDataSize, dwFlagsKF, 
			&lSamplesWritten, &lBytesWritten) ;
	if ( frc )
	{
		fprintf( stderr, "\navi_store_picyv12() : AVIStreamWrite() failed!");
	}
	else
	{
		if ( lCompVideoDataSize != lBytesWritten )
			fprintf( stderr, 
				"\n#bytes written by AVIStreamWrite()!=ICSeqCompressFrame()");
		// update global counters 
		g_lFrameCount += lSamplesWritten;	// count of all frames
		g_lByteCount += lBytesWritten;		// count of bytes
		g_lKFCount += (fIsKeyFrame?1:0);	// count of key-frames
		printf("AVIStreamWrite() (%ld) bytesWritten=%ld %s",
			g_lByteCount, lBytesWritten, (fIsKeyFrame ? "KEYF  ":"      ") );
	}
}